# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Component edit utility functions."""
from typing import Any, Dict, Optional, Tuple, Union

import ipywidgets as widgets
import requests

from ..common.azure_auth_core import AzureCloudConfig
from .comp_edit import SettingsControl
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

_DEBUG = False

ITEM_LIST_LAYOUT = {
    "layout": widgets.Layout(height="150px", width="300px"),
    "style": {"description_width": "70px"},
}

SETTING_LIST_LAYOUT = {
    "layout": widgets.Layout(height="150px", width="300px"),
    "style": {"description_width": "100px"},
}

_TEXT_LAYOUT = {
    "layout": widgets.Layout(width="70%"),
    "style": {"description_width": "100px"},
}

_TEXT_AREA_LAYOUT = {
    "layout": widgets.Layout(width="70%"),
    "style": {"description_width": "100px"},
}


if _DEBUG:

    def print_debug(*args):
        """Print args to std_out (Debug version)."""
        if not _DEBUG:
            return
        print(*args)

else:

    def print_debug(*args):
        """Print nothing std_out (Prod version)."""
        del args


_TEXT_WIDGETS = (widgets.Text, widgets.Textarea, widgets.Label, widgets.Select)


# pylint: disable=too-many-return-statements
def py_to_widget(
    value: Any, ctrl: Optional[widgets.Widget] = None, val_type: Optional[str] = None
) -> Any:
    """
    Adjust type and format to suit target widget.

    Parameters
    ----------
    value : Any
        The value to process
    ctrl : Optional[widgets.Widget], optional
        The target widget type, by default None
    val_type : Optional[str], optional
        The target value type ("str", "bool"), by default None

    Returns
    -------
    Any
        The converted value

    Raises
    ------
    ValueError
        If neither a target control or expected val_type
        are specified.

    Notes
    -----
    This function handles conversion of None to an empty string
    or bools expressed as text strings into actual bools.

    """
    if ctrl is None and val_type is None:
        raise ValueError("Must specify either a target control or expected val_type.")
    if (
        isinstance(ctrl, widgets.Checkbox)
        or val_type == "bool"
        or isinstance(value, bool)
    ):
        if isinstance(value, str):
            return value.casefold() == "true"
        return bool(value)
    if val_type == "txt_dict" or (
        isinstance(ctrl, _TEXT_WIDGETS) and getattr(ctrl, "tag", None) == "txt_dict"
    ):
        return dict_to_txt(value)
    if val_type == "list" or (
        isinstance(ctrl, _TEXT_WIDGETS) and getattr(ctrl, "tag", None) == "list"
    ):
        return "\n".join(value)
    if val_type == "str" or isinstance(ctrl, _TEXT_WIDGETS) or isinstance(value, str):
        if value is None:
            return ""
        return str(value)
    return value


def widget_to_py(ctrl: Union[widgets.Widget, SettingsControl]) -> Any:
    """
    Adjust type and format of value returned from `ctrl.value`.

    Parameters
    ----------
    ctrl : Union[widgets.Widget, SettingsControl]
        The source widget

    Returns
    -------
    Any
        Converted value.

    Notes
    -----
    This function handles conversion of widget values to
    configuration (Python) values.

    """
    if isinstance(ctrl, widgets.Checkbox):
        return ctrl.value
    if isinstance(ctrl, widgets.Textarea) and getattr(ctrl, "tag", None) == "txt_dict":
        return txt_to_dict(ctrl.value) or None
    if isinstance(ctrl, widgets.Textarea) and getattr(ctrl, "tag", None) == "list":
        return ctrl.value.split("\n") if ctrl.value else []
    if isinstance(ctrl, _TEXT_WIDGETS):
        if ctrl.value == "":
            return None
        return str(ctrl.value)
    if isinstance(ctrl, widgets.SelectMultiple):
        return list(ctrl.value)
    return None if not ctrl.value else ctrl.value


# pylint: enable=too-many-return-statements


def get_def_tenant_id(sub_id: str) -> Optional[str]:
    """
    Get the tenant ID for a subscription.

    Parameters
    ----------
    sub_id : str
        Subscription ID

    Returns
    -------
    Optional[str]
        TenantID or None if it could not be found.

    Notes
    -----
    This function returns the tenant ID that owns the subscription.
    This may not be the correct ID to use if you are using delegated
    authorization via Azure Lighthouse.

    """
    res_mgmt_uri = AzureCloudConfig().endpoints.resource_manager
    get_tenant_url = (
        f"{res_mgmt_uri}/subscriptions/{{subscriptionid}}" + "?api-version=2015-01-01"
    )
    resp = requests.get(get_tenant_url.format(subscriptionid=sub_id))
    # Tenant ID is returned in the WWW-Authenticate header/Bearer authorization_uri
    www_header = resp.headers.get("WWW-Authenticate")
    if not www_header:
        return None
    hdr_dict = {
        item.split("=")[0]: item.split("=")[1].strip('"')
        for item in www_header.split(", ")
    }
    tenant_path = hdr_dict.get("Bearer authorization_uri", "").split("/")
    return tenant_path[-1] if tenant_path else None


def txt_to_dict(txt_val: str) -> Dict[str, Any]:
    """
    Return dict from string of "key:val; key2:val2" pairs.

    Parameters
    ----------
    txt_val : str
        The key/value string (items separated by ";",
        key/value separated by ":")

    Returns
    -------
    Dict[str, Any]
        Dictionary of key/values

    """
    if not txt_val:
        return {}
    kvpairs = [
        kv_pair.strip().split(":", maxsplit=1)
        for kv_pair in txt_val.split("\n")
        if kv_pair.strip()
    ]
    return {
        kval[0].strip(): kval[1].strip() if len(kval) > 1 else None for kval in kvpairs
    }


def dict_to_txt(dict_val: Union[str, Dict[str, Any]]) -> str:
    """
    Return string as "key:val; key2:val2" pairs from `dict_val`.

    Parameters
    ----------
    dict_val : Union[str, Dict[str, Any]]
        Dict of key/val pairs
        or string of single key/value

    Returns
    -------
    str
        str formatted as "key:val; key2:val2"

    """
    if isinstance(dict_val, str):
        if not dict_val:
            return ""
        if ":" in dict_val:
            key, val = dict_val.split(":", maxsplit=1)
        else:
            key, val = dict_val, ""
        return f"{key}:{val}"
    if isinstance(dict_val, dict):
        return "\n".join(f"{key}:{val}" for key, val in dict_val.items())
    return ""


# flake8: noqa: F821
def get_wgt_ctrl(
    setting_path: str,
    var_name: str,
    mp_controls: "MpConfigControls",  # type: ignore
    wgt_style: Optional[Dict[str, Any]] = None,
) -> widgets.Widget:
    """
    Return widget appropriate to value type of `var_name`.

    Parameters
    ----------
    setting_path : str
        The setting path (parent path) as dotted string.
    var_name : str
        The key name for the setting below `setting_path`.
    mp_controls : MpConfigControls
        Instance of MpConfigControls data
    wgt_style : Optional[Dict[str, Any]]
        Dict of style and layout items:
        .. parsed-literal::

            {
                "style": {"description_width": "100px"},
                "layout": widgets.Layout(width="50%")
            }

    Returns
    -------
    widgets.Widget
        The widget.

    """
    if wgt_style is None:
        wgt_style = {}
    var_path = f"{setting_path}.{var_name}"
    ctrl = mp_controls.get_control(var_path)
    comp_defn = mp_controls.get_defn(var_path)
    if comp_defn and not isinstance(comp_defn, tuple):
        # definition is a literal
        def_value = comp_defn
    else:
        st_type, st_opts = get_defn_or_default(comp_defn)
        def_value = st_opts["default"] if st_opts and "default" in st_opts else ""
    curr_val = mp_controls.get_value(var_path) or def_value

    if ctrl is None:
        st_type, st_opts = get_defn_or_default(comp_defn)
        if st_type == "bool":
            ctrl = widgets.Checkbox(
                description=var_name,
                value=py_to_widget(curr_val, val_type=st_type),
                **(wgt_style or _TEXT_LAYOUT),
            )
        elif st_type == "enum":
            ctrl = widgets.Select(
                description=var_name,
                options=st_opts.get("options"),
                value=curr_val or "",
                **(wgt_style or SETTING_LIST_LAYOUT),
            )
        elif st_type == "m_enum":
            ctrl = widgets.SelectMultiple(
                description=var_name,
                options=st_opts.get("options"),
                value=curr_val or [],
                **(wgt_style or SETTING_LIST_LAYOUT),
            )
        elif st_type == "txt_dict":
            ctrl = widgets.Textarea(
                description=var_name, value=dict_to_txt(curr_val) or "", **wgt_style
            )
            setattr(ctrl, "tag", "txt_dict")
        elif st_type == "list":
            ctrl = widgets.Textarea(
                description=var_name,
                value=py_to_widget(curr_val, val_type="list") or "",
                **(wgt_style or _TEXT_AREA_LAYOUT),
                # tooltip="Enter each item as 'key:value'. Separate items with new lines.",
            )
            setattr(ctrl, "tag", "list")
        else:
            ctrl = widgets.Text(
                description=var_name,
                value=py_to_widget(curr_val, val_type=st_type),
                **(wgt_style or _TEXT_AREA_LAYOUT),
                # tooltip="Enter items (no quotes) on separate lines.",
            )
        mp_controls.set_control(var_path, ctrl)
    else:
        ctrl.value = py_to_widget(curr_val, ctrl=ctrl)
    return ctrl


def get_defn_or_default(defn: Union[Tuple[str, Any], Any]) -> Tuple[str, Dict]:
    """
    Return the type and options (or a default) for the setting definition.

    Parameters
    ----------
    defn : Optional[Tuple[str, dict]]
        Setting definition. Returns a default of "str", {}
        if no definition is passed.

    Returns
    -------
    Tuple[str, Dict]
        Tuple of setting type and options.

    """
    if isinstance(defn, tuple):
        return defn[0], defn[1]
    return "str", {}


# flake8: noqa: F821
def get_or_create_mpc_section(
    mp_controls: "MpConfigControls", section: str, subkey: Optional[str] = None  # type: ignore
) -> Any:
    """
    Return (and create if it doesn't exist) a settings section.

    Parameters
    ----------
    mp_controls : MpConfigControls
        The MP Config database.
    section : str
        The section name (top level settings item)
    subkey : Optional[str], optional
        Optional subkey to create, by default None

    Returns
    -------
    Any
        The settings at that section[subkey] location.

    """
    curr_section = mp_controls.get_value(section)
    if curr_section is None:
        mp_controls.set_value(section, {})
        curr_section = mp_controls.get_value(section)
    if subkey and subkey not in curr_section:
        mp_controls.set_value(f"{section}.{subkey}", {})
        return mp_controls.get_value(f"{section}.{subkey}")
    return mp_controls.get_value(section)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Compound control classes."""
import os
from copy import deepcopy
from typing import Any, Dict, Optional, Tuple, Union

import ipywidgets as widgets

from .._version import VERSION

try:
    from ..common.keyvault_client import BHKeyVaultClient, MsticpyKeyVaultConfigError
    from ..common.secret_settings import SecretsClient

    _KEYVAULT = True
except ImportError:
    _KEYVAULT = False
from .ce_common import print_debug, py_to_widget, widget_to_py
from .comp_edit import CompEditDisplayMixin, CompEditStatusMixin, SettingsControl
from .mp_config_control import MpConfigControls

__version__ = VERSION
__author__ = "Ian Hellen"


STORE_TEXT = "Text"
STORE_ENV_VAR = "EnvironmentVar"
STORE_KEYVAULT = "KeyVault"


class ArgControl(SettingsControl, CompEditStatusMixin):
    """Args setting element edit component."""

    def __init__(
        self,
        setting_path: Optional[str],
        name: str,
        store_type: str = STORE_TEXT,
        item_value: Any = None,
    ):
        """
        Initialize and ArgControl object.

        Parameters
        ----------
        setting_path : Optional[str], optional
            The full path to the setting (minus the name)
        name : str
            The name of the setting
        store_type : str, optional
            The storage type for the setting value, by default "Text"
            Other options are "EnvironmentVar" and "KeyVault"
        item_value : Any, optional
            The value of the setting, by default None
            Note None is a legitimate value for store_type "KeyVault"

        """
        self.setting_path = setting_path
        self.name = name
        self.kv_client = None

        self.lbl_setting = widgets.Label(
            value=self.name, layout=widgets.Layout(width="130px")
        )
        self.rb_store_type = widgets.RadioButtons(
            options=[STORE_TEXT, STORE_ENV_VAR, STORE_KEYVAULT],
            description="Storage:",
            layout=widgets.Layout(width="35%"),
            style={"description_width": "100px"},
        )
        self.rb_store_type.value = store_type
        if isinstance(item_value, bool):
            self.txt_val = widgets.Checkbox(description="Value")
        else:
            if item_value is None:
                item_value = ""
            self.txt_val = widgets.Text(
                description="Value" if store_type == STORE_TEXT else store_type,
                layout=widgets.Layout(width="99%"),
                style={"description_width": "100px"},
                value=str(item_value) or "",
            )
        # KeyVault settings
        self.cb_kv_def = widgets.Checkbox(description="Def KV Path")
        self.btn_add_kv_secret = widgets.Button(description="Upload to KV")
        self.btn_add_kv_secret.on_click(self._set_kv_secret)
        self.cb_kv_def.value = not bool(item_value)
        self._set_kv_visibility()
        contents_hbox = widgets.HBox(
            [
                self.rb_store_type,
                widgets.VBox(
                    [self.txt_val, self.cb_kv_def, self.btn_add_kv_secret],
                    layout=widgets.Layout(width="64%"),
                ),
            ]
        )
        self.hbox = widgets.VBox(
            [self.lbl_setting, contents_hbox, self.status],
            layout=CompEditDisplayMixin.no_border_layout("99%"),
        )

        self.cb_kv_def.observe(self._disable_txt, names="value")
        self.rb_store_type.observe(self._change_store, names="value")

    @property
    def value(self) -> Union[str, Dict[str, Optional[str]]]:
        """
        Return the value of the control.

        Returns
        -------
        Union[str, Dict[str, Optional[str]]]
            Either a string value or a dict of:
            {"EnvironmentVar": value} or
            {"KeyVault": value or None}

        """
        if self.rb_store_type.value == STORE_TEXT:
            return widget_to_py(self.txt_val)
        return {self.rb_store_type.value: widget_to_py(self.txt_val)}

    @value.setter
    def value(self, value: Union[str, Dict[str, Optional[str]]]):
        """
        Set control to value.

        Parameters
        ----------
        value : Union[str, Dict[str, Optional[str]]]
            Either a str or a dict of:
            .. parsed_literal::

                {"Text": value} or
                {"EnvironmentVar": value} or
                {"KeyVault": value or None}

        """
        print_debug("_ArgControl setter val:", value)
        if not isinstance(value, dict):
            self.txt_val.value = py_to_widget(value, val_type="str")
            self.rb_store_type.value = STORE_TEXT
            return
        store_type, store_val = next(iter(value.items()))
        self.txt_val.value = py_to_widget(store_val, val_type="str")
        self.rb_store_type.value = store_type
        self.cb_kv_def.value = not bool(store_val)
        self._set_kv_visibility()

    def _change_store(self, change):
        """Handle event for store type radio button."""
        st_type = change.get("new")
        self.txt_val.description = "Value" if st_type == STORE_TEXT else st_type
        self._set_kv_visibility()
        self.lbl_setting.value = st_type
        if st_type == STORE_KEYVAULT and not self.txt_val.value:
            self.cb_kv_def.value = True

    def _disable_txt(self, change):
        """Disable the text field if KeyVault and kv_def_enabled."""
        if self.rb_store_type.value != STORE_KEYVAULT:
            return
        kv_def_enabled = change.get("new")
        if kv_def_enabled:
            self.txt_val.value = ""
        self.txt_val.disabled = kv_def_enabled

    def _set_kv_visibility(self):
        """Set the visibility of the keyvault-related controls."""
        if self.rb_store_type.value == STORE_KEYVAULT:
            self.cb_kv_def.layout.visibility = "visible"
            self.btn_add_kv_secret.layout.visibility = "hidden"
        else:
            self.cb_kv_def.layout.visibility = "hidden"
            self.btn_add_kv_secret.layout.visibility = "visible"

    def _set_kv_secret(self, btn):
        """Try to store the current value to key vault."""
        del btn
        if not self.setting_path:
            self.set_status("No setting path to create KV secret name.")
            return
        sec_value = None
        if self.rb_store_type.value == STORE_TEXT:
            sec_value = self.txt_val.value
        elif self.rb_store_type.value == STORE_ENV_VAR:
            sec_value = os.environ.get(self.txt_val.value.strip())
        else:
            return
        if not sec_value:
            self.set_status("No secret value to store.")
            return
        kv_client = self.kv_client
        if _KEYVAULT:
            result, status, kv_client = _set_kv_secret_value(
                setting_path=self.setting_path,
                item_name=self.name,
                value=sec_value,
                kv_client=kv_client,
            )
        else:
            self.set_status("Azure keyvault support is not installed.")
            return
        if result:
            self.set_status(status)
            self.kv_client = kv_client
            self.rb_store_type.value = STORE_KEYVAULT
            self.txt_val.value = ""
            self.cb_kv_def.value = True
        else:
            self.set_status("Error saving secret: status")


def _get_args_val(arg_setting):
    """Return a dict whether the value is a str or a dict."""
    _, arg_val = next(iter(arg_setting.items()))
    if isinstance(arg_val, (str, int, bool)):
        return STORE_TEXT, arg_val
    return next(iter(arg_val.items()))


def get_arg_ctrl(setting_path, var_name, mp_controls):
    """Create the ArgControl based on the current value of the setting."""
    var_path = f"{setting_path}.{var_name}"
    arg_ctrl = mp_controls.get_control(var_path)
    curr_val = mp_controls.get_value(var_path)

    print_debug(var_name)
    if arg_ctrl is None:
        if curr_val is not None:
            store_type, store_val = _get_args_val({var_name: curr_val})
        else:
            store_type, store_val = STORE_TEXT, ""
        arg_ctrl = ArgControl(
            setting_path=setting_path,
            name=var_name,
            store_type=store_type,
            item_value=store_val,
        )
        mp_controls.set_control(var_path, arg_ctrl)
    else:
        arg_ctrl.value = curr_val
    return arg_ctrl


if _KEYVAULT:

    def _set_kv_secret_value(
        setting_path: str,
        item_name: str,
        value: str,
        kv_client: Any = None,
    ) -> Tuple[bool, str, Any]:
        """
        Set the Key Vault secret to `value`.

        Parameters
        ----------
        setting_path : str
            The setting path (parent path)
        item_name : str
            The setting name
        value : str
            The value to store
        kv_client : Optional[BHKeyVaultClient], optional
            Cached Key Vault secrets client if one is already
            attached to the ArgControl, by default None

        Returns
        -------
        Tuple[bool, str, Optional[BHKeyVaultClient]]
            True - if success
            Status - status string
            KeyVault secrets client (used to cache the client)

        """
        try:
            secret_name = SecretsClient.format_kv_name(f"{setting_path}.{item_name}")
            kv_client = kv_client or BHKeyVaultClient()
            kv_client.set_secret(secret_name, value)
            return True, f"Saved as {secret_name}", kv_client
        except MsticpyKeyVaultConfigError:
            return False, "Missing or invalid Key Vault configuration", None
        except Exception as err:  # pylint: disable=broad-except
            return False, str(err), None

else:

    def _set_kv_secret_value(
        setting_path: str,
        item_name: str,
        value: str,
        kv_client: Any = None,
    ) -> Tuple[bool, str, Any]:
        """Return empty response function if Key Vault cannot be initialized."""
        del setting_path, item_name, value, kv_client
        return False, "Azure keyvault libraries are not installed", None


class UserDefQryProvCtrl(SettingsControl):
    """User Defaults Query Provider edit component."""

    def __init__(self, prov_name: str):
        """
        Initialize the control.

        Parameters
        ----------
        prov_name : str
            The query provider name

        """
        self.prov_name = prov_name
        self.prov_type = "Workspace"
        self._set_prov_name(prov_name)
        w_style = {"description_width": "100px"}

        self.txt_alias = widgets.Text(description="Alias", style=w_style)
        self.cb_connect = widgets.Checkbox(description="Auto-connect", style=w_style)
        self.layout = widgets.VBox(
            [self.lbl_type, widgets.VBox([self.txt_alias, self.cb_connect])]
        )

    def _set_prov_name(self, prov_name):
        """
        Set the provider name.

        Notes
        -----
        The provider name can be a simple string or a compound,
        dotted string (in the case of AzureSentinel it will be of the form
        AzureSentinel.WorkspaceName)

        """
        self.prov_name = prov_name
        self.prov_type = "Workspace" if "." in prov_name else "Provider"
        self.lbl_type = widgets.Label(
            value=f"{self.prov_name} ({self.prov_type})",
            layout=widgets.Layout(width="300px"),
        )

    @property
    def value(self) -> Union[str, Dict[str, Optional[str]]]:
        """
        Return the current value of the control.

        Returns
        -------
        Union[str, Dict[str, Optional[str]]]
            The value dict.
            In cases where optional 'alias' and 'connect' settings
            are not used this will be an empty dictionary.

        """
        alias = {"alias": self.txt_alias.value} if self.txt_alias.value else {}
        connect = (
            {"connect": self.cb_connect.value} if not self.cb_connect.value else {}
        )
        return {**alias, **connect}

    @value.setter
    def value(self, value: Union[str, Dict[str, Optional[str]]]):
        """
        Set the value of the component from settings.

        Parameters
        ----------
        value : Optional[Dict[str, str]]
            The current value to set in the control.

        """
        if not value or isinstance(value, str):
            value = {}
        self.txt_alias.value = value.get("alias", "")
        self.cb_connect.value = value.get("connect", True)


class UserDefLoadComponent(SettingsControl):
    """User Defaults Load component edit component."""

    _W_STYLE = {"description_width": "100px"}

    # pylint: disable=line-too-long
    def __init__(
        self, mp_controls: MpConfigControls, comp_name: str, setting_path: str
    ):
        """
        Initialize the control.

        Parameters
        ----------
        mp_controls : MpConfigControls
            Msticpy configu controls data store.
        comp_name : str
            Component name
        setting_path : str
            Path to setting (minus comp_name)

        """
        # pylint: enable=line-too-long
        self.comp_name = comp_name

        self.comp_path = f"{setting_path}.{comp_name}"
        self.control_map = deepcopy(mp_controls.get_defn(self.comp_path))
        self.controls = {}
        self.controls["label"] = widgets.Label(value=self.comp_name)

        # build controls from dict
        self._create_controls(self.comp_path, mp_controls)

    def _create_controls(self, path, mp_controls):
        """Create controls from settings."""
        comp_defn = mp_controls.get_defn(path)
        # If the setting is a simple string - no configurable elements
        if comp_defn is None:
            self.controls["no_settings"] = widgets.Label(
                value="No settings for this component."
            )
            return

        if not isinstance(comp_defn, dict):
            raise TypeError(f"component definition type is invalid {type(comp_defn)}")

        for name, settings in comp_defn.items():
            curr_value = mp_controls.get_value(f"{path}.{name}")
            if settings is None:
                self.controls["no_settings"] = widgets.Label(
                    value="No settings for this component."
                )
                return
            ctrl_path = f"{path}.{name}"
            if isinstance(settings, str):
                # Simple case of a string value
                self.controls[name] = widgets.Text(
                    description="Value", value=curr_value or ""
                )
                self._add_control_to_map(ctrl_path, self.controls[name])
            if isinstance(settings, tuple):
                # if tuple then the second elem of the tuple is the type defn
                self.controls[name] = self._create_select_ctrl(
                    settings, name, curr_value
                )
                self._add_control_to_map(ctrl_path, self.controls[name])
            elif isinstance(settings, dict):
                self.controls[name] = widgets.Text(value=name, disabled=True)
                self._create_controls(ctrl_path, mp_controls)

    def _add_control_to_map(self, path, ctrl):
        """Set a value at dotted path location in self.control_map dict."""
        ctrl_path = path.replace(f"{self.comp_path}.", "")
        ctrl_map = self.control_map
        for elem in ctrl_path.split("."):
            if not isinstance(ctrl_map.get(elem), dict):
                ctrl_map[elem] = ctrl
                break
            ctrl_map = ctrl_map.get(elem)

    @property
    def value(self) -> Union[str, Dict[str, Optional[str]]]:
        """
        Return the current value of the control.

        Returns
        -------
        Optional[Dict[str, Any]]
            Control value dictionary.

        """
        return self._get_val_from_ctrl(self.control_map)

    @value.setter
    def value(self, value: Union[str, Dict[str, Optional[str]]]):
        """Set value of controls from dict."""
        if isinstance(value, dict):
            self._set_ctrl_from_val(path="", value=value)

    def _get_val_from_ctrl(self, val_dict):
        """Recursive get values from control dictionary."""
        ctrl_val = {}
        if not val_dict:
            return val_dict
        for name, value in val_dict.items():
            if isinstance(value, widgets.Label):
                continue
            if isinstance(value, dict):
                ctrl_val[name] = self._get_val_from_ctrl(value)
            elif isinstance(value, widgets.Widget):
                ctrl_val[name] = widget_to_py(ctrl=value)
            else:
                ctrl_val[name] = value
        return ctrl_val

    def _set_ctrl_from_val(self, path, value):
        if value is None:
            return
        for key, val in value.items():
            sub_path = f"{path}.{key}" if path else key
            if isinstance(val, dict):
                if isinstance(self.controls[key], widgets.Textarea):
                    self.controls[key].value = py_to_widget(
                        val, ctrl=self.controls[key]
                    )
                else:
                    self._set_ctrl_from_val(sub_path, val)
            elif key in self.controls:
                self.controls[key].value = py_to_widget(val, ctrl=self.controls[key])

    @property
    def layout(self):
        """Return the widget layout for the control."""
        return widgets.VBox(list(self.controls.values()))

    def _create_select_ctrl(self, ctrl_defn, name, curr_value):
        val_type, val_opts = ctrl_defn
        if curr_value is None:
            curr_value = val_opts.get("default")

        if val_type == "str":
            return widgets.Text(
                description=name,
                value=py_to_widget(curr_value, val_type=val_type),
                style=self._W_STYLE,
            )
        if val_type == "bool":
            if isinstance(curr_value, str):
                curr_value = curr_value.casefold() == "true"
            return widgets.Checkbox(
                description=name,
                value=py_to_widget(curr_value, val_type=val_type),
                style=self._W_STYLE,
            )
        if val_type in ("enum", "m_enum"):
            return widgets.SelectMultiple(
                description=name,
                options=val_opts.get("options"),
                value=curr_value or [],
                style=self._W_STYLE,
            )
        if val_type == "txt_dict":
            wgt = widgets.Textarea(
                description=name,
                style=self._W_STYLE,
            )
            setattr(wgt, "tag", "txt_dict")
            wgt.value = py_to_widget(curr_value, ctrl=wgt) or ""
            return wgt
        raise TypeError(f"Unknown definition type {val_type} for {name}/{ctrl_defn}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module docstring."""
from datetime import datetime

import ipywidgets as widgets

from .._version import VERSION
from .ce_common import ITEM_LIST_LAYOUT, print_debug
from .ce_common import get_wgt_ctrl, get_or_create_mpc_section
from .comp_edit import CEItemsBase, CompEditDisplayMixin
from .mp_config_control import MpConfigControls

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-ancestors
class CEAzureSentinel(CEItemsBase):
    """Microsoft Sentinel Workspaces editor component."""

    _DESCRIPTION = "Microsoft Sentinel workspace settings"
    _COMP_PATH = "AzureSentinel.Workspaces"
    _HELP_TEXT = """
    Supply the parameters for your Microsoft Sentinel workspaces here.<br>

    You can get all of these (apart from 'TenantID') from your workspace portal.
    Navigate to "Settings" (on the left side of the screen), then click the
    "Workspace Settings" tab near the top of the page.

    The Tenant ID is the only value missing from this settings page. If you cannot
    find this you can use the following code to try to look it up:
    <pre>
    from msticpy.common.config.ce_common import get_def_tenant_id
    get_def_tenant_id("{workspace-subscription-id}")
    </pre>

    The name that you use for workspace (the "Name" text box) does not have to
    be the same as the official name. You can use any helpful name that you
    like. You can even have the same workspace included multiple times with
    different names.

    There is a special case of the workspace name "Default". Normally, when you
    connect to a workspace you need to supply the name corresponding to the
    workspace entry in your settings. If you give a workspace the name "Default"
    it will be picked if you do not specify a name. Note, the "Default" entry
    can be a duplicate of another named entry.
    """
    _HELP_URI = {
        "Connecting to a Microsoft Sentinel Workspace": (
            "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
            + "DataProviders.html#connecting-to-an-azure-sentinel-workspace"
        ),
        "Help on this tab": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "SettingsEditor.html#azure-sentinel-workspaces"
        ),
    }

    def __init__(self, mp_controls: MpConfigControls):
        """
        Initialize an instance of CEAzureSentinel.

        Parameters
        ----------
        mp_controls : MpConfigControls
            The config/controls/settings database

        """
        super().__init__(mp_controls)

        get_or_create_mpc_section(self.mp_controls, self._COMP_PATH)
        self.select_item.options = self._get_select_opts()
        self.select_item.layout = ITEM_LIST_LAYOUT["layout"]
        self.select_item.style = ITEM_LIST_LAYOUT["style"]
        self.select_item.description = "Workspaces"
        self.select_item.observe(self._select_provider, names="label")

        prov_name = self.select_item.label
        self.edit_ctrls = _get_ws_ctrls(prov_name, self.mp_controls, self._COMP_PATH)
        self.edit_ctrls.children[0].value = prov_name or ""
        self.btn_set_default = widgets.Button(description="Set as default")
        if prov_name:
            self.edit_frame.children = [self.edit_ctrls, self.btn_set_default]
        else:
            self.edit_frame.children = [self.edit_ctrls]
        self.edit_buttons.btn_del.on_click(self._del_item)
        self.edit_buttons.btn_add.on_click(self._add_item)
        self.edit_buttons.btn_save.on_click(self._save_item)
        self.btn_set_default.on_click(self._set_default)

        self.current_workspace = prov_name

    @property
    def _current_path(self):
        return f"{self._COMP_PATH}.{self.select_item.label}"

    def _get_select_opts(self):
        """Get provider options to populate select list."""
        provs = self.mp_controls.get_value(self._COMP_PATH)
        return [(val, idx) for idx, val in enumerate(provs.keys())]

    def _select_provider(self, change):
        """Select a provider from the list."""
        prov_name = change.get("new")
        self.edit_ctrls = _get_ws_ctrls(prov_name, self.mp_controls, self._COMP_PATH)
        self.edit_ctrls.children[0].value = prov_name
        self.current_workspace = prov_name
        self.edit_frame.children = [self.edit_ctrls, self.btn_set_default]
        self.mp_controls.populate_ctrl_values(self.current_workspace)

    def _add_item(self, btn):
        """Add an item."""
        del btn
        # Generate a "unique" temp name
        new_provider = f"NewWorkspace_{datetime.now().strftime('%f')[:4]}"
        while new_provider in dict(self.select_item.options):
            new_provider = f"NewWorkspace_{datetime.now().strftime('%f')[:4]}"
        _get_ws_ctrls(new_provider, self.mp_controls, self._COMP_PATH)
        self.mp_controls.save_ctrl_values(f"{self._COMP_PATH}.{new_provider}")
        current_options = self._get_select_opts()
        self.select_item.options = current_options
        self.select_item.label = new_provider

    def _del_item(self, btn):
        """Delete an item."""
        del btn
        self.mp_controls.del_value(self._current_path)
        remaining_opts = self._get_select_opts()
        self.select_item.options = remaining_opts
        if remaining_opts:
            self.select_item.label = remaining_opts[-1][0]

    def _save_item(self, btn):
        """Save the current item."""
        del btn
        # Use may have edited the workspace name.
        # If so save the existing one and rename the paths to match new name
        edited_provider_name = self.edit_ctrls.children[0].value
        self.mp_controls.save_ctrl_values(self._current_path)
        if edited_provider_name != self.current_workspace and self.current_workspace:
            self.mp_controls.rename_path(
                f"{self._COMP_PATH}.{self.current_workspace}",
                f"{self._COMP_PATH}.{edited_provider_name}",
            )
            self.select_item.options = self._get_select_opts()
            self.select_item.label = edited_provider_name
        valid, status = _validate_ws(
            edited_provider_name, self.mp_controls, self._COMP_PATH
        )
        if not valid:
            self.set_status(status)

    def _set_default(self, btn):
        """Set selected item to be the Default entry."""
        del btn
        src_wkspace = self.current_workspace
        if src_wkspace == "Default" or not src_wkspace:
            return
        src_path = f"{self._COMP_PATH}.{src_wkspace}"
        src_settings = self.mp_controls.get_value(src_path)
        def_path = f"{self._COMP_PATH}.Default"
        # save the src values to the Default key and refresh the select_item list
        self.mp_controls.set_value(def_path, src_settings)
        print_debug("Current options", self._get_select_opts())
        self.select_item.options = self._get_select_opts()
        self.select_item.label = "Default"

    def _select_labels(self):
        return [label for label, _ in self.select_item.options]


_TEXT_PARAMS = {
    "layout": widgets.Layout(width="70%"),
    "style": {"description_width": "150px"},
}


def _get_ws_ctrls(workspace, mp_controls, conf_path):
    """Return or create the control set for a workspace."""
    defn_path = f"{conf_path}.Default"
    prov_defn = mp_controls.get_defn(defn_path)

    ctrls = [widgets.Text(description="Name", **_TEXT_PARAMS)]
    if not workspace:
        return widgets.VBox(ctrls, layout=CompEditDisplayMixin.no_border_layout("95%"))
    prov_path = f"{conf_path}.{workspace}"
    for setting in prov_defn:
        wgt = get_wgt_ctrl(prov_path, setting, mp_controls)
        wgt.layout = _TEXT_PARAMS["layout"]
        wgt.style = _TEXT_PARAMS["style"]
        ctrls.append(wgt)

    return widgets.VBox(ctrls)


def _validate_ws(workspace, mp_controls, conf_path):
    """Validate the settings for a workspace."""
    defn_path = f"{conf_path}.Default"

    if not workspace:
        return False, "Workspace name must be supplied."
    prov_path = f"{conf_path}.{workspace}"
    results = mp_controls.validate_setting(path=prov_path, defn_path=defn_path)
    status = []
    validated = True
    for valid, mssg in results:
        if not valid:
            validated = False
            status.append(mssg)
    return validated, "  ".join(status)


#
#
# _KNOWN_SEC_NAMES = ("AuthKey", "ApiID", "clientSecret", "password")


# def _enum_template(template: Dict[str, Any]):
#     output: Dict[str, Any] = {}
#     for key, value in template.items():
#         if key in _KNOWN_SEC_NAMES:
#             output[key] = ""
#         elif isinstance(value, dict):
#             output[key] = _enum_template(value)
#         elif isinstance(value, tuple):
#             val_type, val_opt = value
#             if val_type == "str":
#                 output[key] = val_opt.get("default", "")
#             elif val_type == "bool":
#                 bool_str = val_opt.get("default", False)
#                 if isinstance(bool_str, str):
#                     output[key] = bool_str.casefold() == "true"
#                 else:
#                     output[key] = bool_str
#             elif val_type == "int":
#                 output[key] = int(val_opt.get("default", 0))
#             elif val_type == "enum":
#                 output[key] = val_opt.get("default", val_opt.get("options")[0])
#         else:
#             output[key] = value
#     return output



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Component Edit base and mixin classes."""
from abc import ABC, abstractmethod
from time import sleep
from typing import Any, Dict, List, Optional, Tuple, Union

import ipywidgets as widgets
from IPython.display import display

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


class CompEditDisplayMixin:
    """Mixin class with common display methods."""

    def _ipython_display_(self):
        layout = getattr(self, "layout", None)
        if layout:
            display(layout)

    @staticmethod
    def border_layout(width="95%"):
        """Return border widget layout."""
        return widgets.Layout(
            **{
                "width": width,
                "border": "solid gray 1px",
                "margin": "1pt",
                "padding": "5pt",
            }
        )

    @staticmethod
    def no_border_layout(width="95%"):
        """Return no-border widget layout."""
        return widgets.Layout(**{"width": width, "margin": "1pt", "padding": "5pt"})


# pylint: disable=too-few-public-methods
class CompEditUtilsMixin:
    """Mixin class with common display methods."""

    @staticmethod
    def _get_settings_path(settings, path):
        path_elems = path.split(".")
        curr_node = settings
        for elem in path_elems:
            curr_node = curr_node.get(elem)
            if curr_node is None:
                break
        return curr_node


class CompEditStatusMixin:
    """Mixin class with with status label."""

    testing = False
    status = widgets.Label(layout=widgets.Layout(width="99%"))

    def set_status(self, status, timeout: float = 3.0):
        """Set the status text."""
        self.status.value = status
        if timeout == 0 or self.testing:
            return
        sleep(timeout)
        self.clear_status()

    def clear_status(self):
        """Clear the status text."""
        self.status.value = ""


class CompEditHelp:
    """Class to add help control."""

    _DEFAULT_URI = {
        "MSTICPy Config": (
            "https://msticpy.readthedocs.io/en/latest/"
            + "getting_started/msticpyconfig.html"
        )
    }
    _HELP_STYLE = "color: blue; text-decoration: underline;"

    def __init__(self, help_text: str = "", help_uri: Dict[str, str] = None):
        """
        Create help sub-component.

        Parameters
        ----------
        help_text : str, optional
            The help string (HTML), by default ""
        help_uri : Dict[str, str], optional
            Dict of named URIs {disp_txt: URI}, by default None

        """
        self.html_help = widgets.HTML()
        self.accdn_help = widgets.Accordion(children=[self.html_help])
        self.accdn_help.set_title(0, "Help")
        self.accdn_help.selected_index = None
        self.set_help(help_text, help_uri)

    def set_help(self, help_text: str = "", help_uri: Dict[str, str] = None):
        """Set the help string (HTML) and URIs."""
        if not help_uri:
            help_uri = self._DEFAULT_URI
        help_uris = "<br>".join(
            f"<a href='{uri}' target='_blank' style='{self._HELP_STYLE}'>Online: {title}</a>"
            for title, uri in help_uri.items()
        )
        help_text = "<br>".join(
            [
                help_text,
                help_uris,
            ]
        )
        self.html_help.value = help_text

    @property
    def layout(self):
        """Return the parent widget."""
        return self.accdn_help


class CompEditFrame(CompEditDisplayMixin, CompEditUtilsMixin, CompEditStatusMixin):
    """Edit frame class for components."""

    def __init__(self, description: str = None):
        """Initialize the class. Set a label with `description` as content."""
        self.inner_frame = widgets.HBox(layout=self.border_layout("99%"))
        self.description = widgets.Label(value=description or "")
        # self.description.style = {"visibility": "visible" if description else "hidden"}
        self.help = CompEditHelp()
        # self.status = widgets.Label(layout=widgets.Layout(width="99%"))
        self.layout = widgets.VBox(
            [self.description, self.inner_frame, self.help.layout, self.status]
        )


class CompEditItemButtons:
    """Component class to add default buttons."""

    def __init__(self):
        """Initialize the class."""
        btn_layout = widgets.Layout(width="100px")
        self.btn_add = widgets.Button(
            description="Add",
            tooltip="Add provider/item in drop-down to your settings.",
            layout=btn_layout,
        )
        self.btn_del = widgets.Button(
            description="Delete",
            tooltip="Remove selected provider/item from your settings.",
            layout=btn_layout,
        )
        self.btn_save = widgets.Button(
            description="Update",
            tooltip="Confirms updates to the settings changes",
            layout=btn_layout,
        )
        self.layout = widgets.HBox([self.btn_add, self.btn_del, self.btn_save])


class CompEditItems(CompEditFrame):
    """Base class for item list and edit controls."""

    def __init__(self, description: str):
        """Initialize the class. Set a label with `description` as content."""
        super().__init__(description=description)
        self.select_item = widgets.Select(
            layout=widgets.Layout(height="200px", width="99%")
        )
        self.edit_frame = widgets.VBox(layout=self.border_layout("99%"))
        self.edit_buttons = CompEditItemButtons()
        self.items_frame = widgets.VBox(
            [
                self.select_item,
                widgets.HBox([self.edit_buttons.btn_add, self.edit_buttons.btn_del]),
            ],
            layout=self.no_border_layout(width="40%"),
        )
        self.inner_frame.children = [
            self.items_frame,
            widgets.VBox(
                [self.edit_frame, self.edit_buttons.btn_save],
                layout=widgets.Layout(width="99%"),
            ),
        ]


class CompEditSimple(CompEditFrame):
    """Base class for simple component with only edit controls."""

    def __init__(self, description: str):
        """Initialize the class. Set a label with `description` as content."""
        super().__init__(description=description)
        self.edit_frame = widgets.VBox(layout=self.border_layout("100%"))
        self.btn_save = widgets.Button(
            description="Update", tooltip="Confirms updates to the settings changes"
        )
        self.container = widgets.VBox(
            [self.edit_frame, self.btn_save], layout=self.no_border_layout("100%")
        )
        self.inner_frame.children = [self.container]


class CEItemsBase(CompEditItems, ABC):
    """Base class for components containing an item list."""

    _DESCRIPTION = "No description"
    _COMP_PATH = ""
    _HELP_TEXT = """"""
    _HELP_URI = {
        "MSTICPy Configuration": (
            "https://msticpy.readthedocs.io/en/latest/"
            + "getting_started/msticpyconfig.html"
        )
    }

    def __init__(self, mp_controls):
        """Initialize the class. Set the controls and retrieve settings."""
        super().__init__(description=self._DESCRIPTION)

        self.mp_controls = mp_controls
        self.comp_defn = self.mp_controls.get_defn(self._COMP_PATH)
        self.settings = self.mp_controls.get_value(self._COMP_PATH)

        self.help.set_help(self._HELP_TEXT, self._HELP_URI)

        self.controls = {}


class SettingsControl(ABC):
    """Abstract base class for settings controls."""

    @property
    @abstractmethod
    def value(self) -> Union[str, Dict[str, Optional[str]]]:
        """Return the current value of the control."""

    @value.setter
    def value(self, value: Union[str, Dict[str, Optional[str]]]):
        """Set value of controls from dict."""


CETabControlDef = Tuple[type, Union[List[Any], Dict[str, Any]]]


class CompEditTabs:
    """Tab class."""

    def __init__(self, tabs: Optional[Dict[str, CETabControlDef]] = None):
        """
        Initialize the CompEditTabs class.

        Parameters
        ----------
        tabs : Optional[Dict[str, Tuple[type, Union[List[Any], Dict[str, Any]]]]], optional
            Tab definitions or contents, by default None.
            Each definition can be a Tuple of class and list of args
            or a Tuple of class and dict of kwargs.

        """
        self.tab = widgets.Tab()
        self.layout = self.tab
        tabs = tabs or {}
        self._tab_state: List[widgets.Widget] = []
        self._tab_lazy_load: Dict[int, CETabControlDef] = {}
        self._tab_names: List[str] = []
        self.controls: Dict[str, Any] = {}
        if tabs:
            for tab_name, tab_ctrl in tabs.items():
                if isinstance(tab_ctrl, CEItemsBase):
                    # if this is an already-instantiated widget, just add the tab
                    self.add_tab(tab_name, tab_ctrl)
                    self._tab_state.append(tab_ctrl.layout)
                elif isinstance(tab_ctrl, tuple):
                    # if we're doing lazy loading, add a lazy-load tab
                    self._add_lazy_tab(tab_name, tab_ctrl)
            if self._tab_lazy_load:
                # Make sure content is loaded for tab 0
                self._load_tab(0)
        self.tab.observe(self._on_select_tab, names="selected_index")

    def _load_tab(self, tab_index):
        """Load any lazily-loaded tab content."""
        if not self._tab_lazy_load:
            return
        if isinstance(self._tab_state[tab_index], widgets.Label):
            wgt_cls, args = self._tab_lazy_load[tab_index]
            # initialize the class object with args or kwargs
            ctrl = wgt_cls(*args) if isinstance(args, list) else wgt_cls(**args)
            # set this control as the current tab state
            self._tab_state[tab_index] = ctrl.layout
            # update the controls dict
            self.controls[self._tab_names[tab_index]] = ctrl
            # update the tab control children with the new state
            self.tab.children = self._tab_state

    def _on_select_tab(self, change):
        """Handle tab select index events."""
        tab_index = change.get("new")
        self._load_tab(tab_index)

    def add_tab(self, tab_name: str, control: CEItemsBase):
        """Add a tab with name `tab_name` and content `control`."""
        self._tab_names.append(tab_name)
        new_idx = len(self._tab_state)
        self._tab_state.append(control.layout)
        self.tab.children = self._tab_state
        self.tab.set_title(new_idx, tab_name)
        self.controls[tab_name] = control

    def _add_lazy_tab(self, tab_name: str, control_def: CETabControlDef):
        """Add a lazily-loaded tab with name `tab_name` and definition `control_def`."""
        self._tab_names.append(tab_name)
        new_idx = len(self._tab_state)
        # add dummy control to tab state and control def to lazy tab dict
        dummy_ctrl = widgets.Label(value="loading...")
        self._tab_state.append(dummy_ctrl)
        self._tab_lazy_load[new_idx] = control_def
        self.controls[tab_name] = dummy_ctrl
        # Refresh tab control children and set title
        self.tab.children = self._tab_state
        self.tab.set_title(new_idx, tab_name)

    def set_tab(self, tab_name: Optional[str], index: int = 0):
        """Programatically set the tab by name or index."""
        if tab_name:
            tab_index = [
                idx
                for idx, tabname in enumerate(self._tab_names)
                if tab_name.casefold() == tabname.casefold()
            ]
            if tab_index:
                self.tab.selected_index = tab_index[0]
            return
        self.tab.selected_index = index

    @property
    def tab_names(self) -> List[str]:
        """Return a list of current tabs."""
        return self._tab_names

    @property
    def tab_controls(self) -> Dict[str, Any]:
        """Return a list of current tab names and controls."""
        return self.controls



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Tree Schema module for Process Tree Visualization."""
from typing import Any, Dict, Optional

import attr
import pandas as pd

from .._version import VERSION
from ..common.exceptions import MsticpyUserError

__version__ = VERSION
__author__ = "Ian Hellen"


class ProcessTreeSchemaException(MsticpyUserError):
    """Custom exception for Process Tree schema."""

    DEF_HELP_URI = (
        "MSTICPy Process Tree documentation",
        "https://msticpy.readthedocs.io/en/latest/visualization/ProcessTree.html",
    )


@attr.s(auto_attribs=True)
class ProcSchema:
    """
    Property name lookup for Process event schema.

    Each property maps a generic column name on to the
    schema of the input data. Most of these are mandatory,
    some are optional - not supplying them may result in
    a less complete tree.
    The `time_stamp` column should be supplied although
    defaults to 'TimeGenerated'.

    """

    process_name: str
    process_id: str
    parent_id: str
    logon_id: str
    cmd_line: str
    user_name: str
    path_separator: str
    host_name_column: str
    time_stamp: str = "TimeGenerated"
    parent_name: Optional[str] = None
    target_logon_id: Optional[str] = None
    user_id: Optional[str] = None
    event_id_column: Optional[str] = None
    event_id_identifier: Optional[Any] = None

    def __eq__(self, other):
        """Return False if any non-blank field values are unequal."""
        if not isinstance(other, ProcSchema):
            return False
        self_dict = attr.asdict(self)

        return not any(
            value and value != self_dict[field]
            for field, value in attr.asdict(other).items()
        )

    @property
    def required_columns(self):
        """Return columns required for Init."""
        return [
            "process_name",
            "process_id",
            "parent_id",
            "logon_id",
            "cmd_line",
            "user_name",
            "path_separator",
            "host_name_column",
            "time_stamp",
        ]

    @property
    def column_map(self) -> Dict[str, str]:
        """Return a dictionary that maps fields to schema names."""
        return {
            prop: str(col)
            for prop, col in attr.asdict(self).items()
            if prop not in {"path_separator", "event_id_identifier"}
        }

    @property
    def columns(self):
        """Return list of columns in schema data source."""
        return [
            col
            for prop, col in attr.asdict(self).items()
            if prop not in {"path_separator", "event_id_identifier"}
        ]

    def get_df_cols(self, data: pd.DataFrame):
        """Return the subset of columns that are present in `data`."""
        return [col for col in self.columns if col in data.columns]

    @property
    def event_type_col(self) -> str:
        """
        Return the column name containing the event identifier.

        Returns
        -------
        str
            The name of the event ID column.

        Raises
        ------
        ProcessTreeSchemaException
            If the schema is not known.

        """
        if self.event_id_column:
            return self.event_id_column
        raise ProcessTreeSchemaException(
            "Unknown schema - there is no value for the 'event_id' column."
        )

    @property
    def event_filter(self) -> Any:
        """
        Return the event type/ID to process for the current schema.

        Returns
        -------
        Any
            The value of the event ID to process.

        Raises
        ------
        ProcessTreeSchemaException
            If the schema is not known.

        """
        if self.event_id_identifier:
            return self.event_id_identifier
        raise ProcessTreeSchemaException(
            "Unknown schema - there is no value for the 'event_id_identifier' in the schema."
        )


WIN_EVENT_SCH = ProcSchema(
    time_stamp="TimeGenerated",
    process_name="NewProcessName",
    process_id="NewProcessId",
    parent_name="ParentProcessName",
    parent_id="ProcessId",
    logon_id="SubjectLogonId",
    target_logon_id="TargetLogonId",
    cmd_line="CommandLine",
    user_name="SubjectUserName",
    path_separator="\\",
    user_id="SubjectUserSid",
    event_id_column="EventID",
    event_id_identifier=4688,
    host_name_column="Computer",
)

LX_EVENT_SCH = ProcSchema(
    time_stamp="TimeGenerated",
    process_name="exe",
    process_id="pid",
    parent_name=None,
    parent_id="ppid",
    logon_id="ses",
    target_logon_id=None,
    cmd_line="cmdline",
    user_name="acct",
    path_separator="/",
    user_id="uid",
    event_id_column="EventType",
    event_id_identifier="SYSCALL_EXECVE",
    host_name_column="Computer",
)

MDE_INT_EVENT_SCH = ProcSchema(
    time_stamp="CreatedProcessCreationTime",
    process_name="CreatedProcessName",
    process_id="CreatedProcessId",
    parent_name="ParentProcessName",
    parent_id="CreatedProcessParentId",
    logon_id="InitiatingProcessLogonId",
    target_logon_id="LogonId",
    cmd_line="CreatedProcessCommandLine",
    user_name="CreatedProcessAccountName",
    path_separator="\\",
    user_id="CreatedProcessAccountSid",
    host_name_column="ComputerDnsName",
)

# MDE Public and Sentinel DeviceProcessEvents schema
MDE_EVENT_SCH = ProcSchema(
    time_stamp="Timestamp",
    process_name="FileName",
    process_id="ProcessId",
    parent_name="InitiatingProcessFileName",
    parent_id="InitiatingProcessParentId",
    logon_id="InitiatingProcessLogonId",
    target_logon_id="LogonId",
    cmd_line="ProcessCommandLine",
    user_name="AccountName",
    path_separator="\\",
    user_id="AccountSid",
    host_name_column="DeviceName",
    event_id_column="ActionType",
    event_id_identifier="ProcessCreated",
)

SUPPORTED_SCHEMAS = (
    WIN_EVENT_SCH,
    LX_EVENT_SCH,
    MDE_INT_EVENT_SCH,
    MDE_EVENT_SCH,
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Dataframe viewer."""
from collections import namedtuple
from typing import Dict, List, Union

import ipywidgets as widgets
import pandas as pd
from bokeh.io import push_notebook, show, output_notebook
from bokeh.models import (
    BooleanFilter,
    CDSView,
    ColumnDataSource,
    DataTable,
    DateFormatter,
    TableColumn,
)
from IPython.display import display

from . import nbwidgets

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


FilterExpr = namedtuple("FilterExpr", "column, inv, operator, expr")


# pylint: disable=too-many-instance-attributes
class DataViewer:
    """Data viewer class."""

    _DEF_HEIGHT = 550

    def __init__(
        self, data: pd.DataFrame, selected_cols: List[str] = None, debug=False
    ):
        """
        Initialize the DataViewer class.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame to view
        selected_cols : List[str], optional
            Initial subset of columns to show, by default None (all cols)
        debug : bool
            Output additional debugging info to std out.

        """
        if data.empty:
            raise ValueError("No data available in 'data'")

        output_notebook(hide_banner=True)
        # Drop empty columns
        data = data.dropna(axis="columns", how="all")
        self.cds = ColumnDataSource(data)
        self._columns = _get_cols_from_df(data)
        self._dt_columns = list(self._columns.values())
        self.data = data
        self._debug = debug

        self.nb_handle = None
        self.data_table = DataTable(
            source=self.cds,
            columns=self._dt_columns,
            view=CDSView(source=self.cds),
            height=self._calc_df_height(data),
            width_policy="max",
            auto_edit=True,
            editable=True,
            reorderable=True,
        )

        self.column_chooser = DataTableColumnChooser(data, selected_cols=selected_cols)
        self.data_filter = DataTableFilter(data)
        if selected_cols is not None:
            self._update_columns(btn=None)

        self.column_chooser.apply_button.on_click(self._update_columns)
        self.data_filter.apply_button.on_click(self._apply_filter)

        self.accordion = widgets.Accordion(
            children=[self.column_chooser.layout, self.data_filter.layout]
        )
        self.accordion.set_title(0, "Choose columns")
        self.accordion.set_title(1, "Filter data")
        self.accordion.selected_index = None

        self.layout = self.accordion

    @property
    def filtered_data(self) -> pd.DataFrame:
        """Return filtered dataframe."""
        return self.data_filter.filtered_dataframe[self.column_chooser.selected_columns]

    @property
    def filters(self) -> Dict[str, FilterExpr]:
        """Return current filters as a dict."""
        return self.data_filter.filters

    def import_filters(self, filters: Dict[str, FilterExpr]):
        """
        Import filter set replacing current filters.

        Parameters
        ----------
        filters : Dict[str, FilterExpr]
            dict of filter name, FilterExpr
            FilterExpr is a tuple of:
            column [str], inv [bool], operator [str], expr [str]

        """
        self.data_filter.import_filters(filters)
        self._apply_filter(btn=None)

    def _calc_df_height(self, data):
        df_height = 20 + (len(data) * 20)
        return min(df_height, self._DEF_HEIGHT)

    def show(self):
        """Display the data table control."""
        if self._debug:
            print("_update_data_table")
        self.nb_handle = show(self.data_table, notebook_handle=True)

    def _update_data_table(self):
        if self._debug:
            print("_update_data_table")
            print(self.data_filter.filters)
            print(len(self.filtered_data))
            print(self.filtered_data.iloc[:2])
        if self.nb_handle:
            push_notebook(handle=self.nb_handle)

    def display(self):
        """Display the widget."""
        self.show()
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()

    def _update_columns(self, btn):
        del btn
        self.data_table.columns = self.column_chooser.datatable_columns
        self._update_data_table()

    def _apply_filter(self, btn):
        del btn
        if self._debug:
            print("_apply_filter")
        self.data_table.view = CDSView(
            source=self.cds, filters=[BooleanFilter(self.data_filter.bool_filters)]
        )
        self.data_table.height = self._calc_df_height(
            self.data_filter.filtered_dataframe
        )
        self._update_data_table()


class DataTableColumnChooser:
    """DataTableColumnChooser class."""

    def __init__(self, data, selected_cols=None):
        """Initialize the DataTableColumnChooser class."""
        self.data = data
        self._all_col_names = list(data.columns)
        self._initial_cols = selected_cols or self._all_col_names
        self._col_select = nbwidgets.SelectSubset(
            default_selected=self._initial_cols,
            source_items=self._all_col_names,
            auto_display=False,
        )

        self.apply_button = widgets.Button(description="Apply columns")
        self.layout = widgets.VBox([self._col_select.layout, self.apply_button])

    @property
    def datatable_columns(self):
        """Return a list of Bokeh column definitions for the DataFrame."""
        return list(_get_cols_from_df(self.dataframe_columns).values())

    @property
    def dataframe_columns(self):
        """Return the selected set of DataFrame columns."""
        return self.data[self._reorder_cols(self.selected_columns)]

    def _reorder_cols(self, columns):
        """Return column list in original order."""
        # order the columns as originally specified (or as the DF)
        col_init = [col for col in self._initial_cols if col in columns]
        # If any new columns, add them to the end of the list
        col_init.extend(list(set(columns) - set(col_init)))
        return col_init

    def display(self):
        """Display in IPython."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()

    @property
    def selected_columns(self):
        """Return the selected columns."""
        return self._reorder_cols(self._col_select.selected_items)


def _layout(width, height=None, desc_width=None, **kwargs):
    """Layout creation for widgets."""
    wgt_dict = {}
    lo_dict = {"width": width}
    if height:
        lo_dict["height"] = height
    border = kwargs.pop("border", None)
    if border:
        lo_dict.update(
            {
                "border": "solid gray 1px",
                "margin": "1pt",
                "padding": "5pt",
            }
        )
    wgt_dict["layout"] = widgets.Layout(**lo_dict)
    style_dict = {}
    if desc_width:
        style_dict["description_width"] = desc_width
    if kwargs:
        style_dict.update(kwargs)
    if style_dict:
        wgt_dict["style"] = style_dict
    return wgt_dict


class DataTableFilter:
    """Data filtering class."""

    _OPERATORS = {
        "string": ["==", "contains", "matches", "in", "between", "query"],
        "other": ["==", ">", "<", ">=", "<=", "in", "between", "query"],
    }

    def __init__(self, data: pd.DataFrame):
        """Initialize the DataTableFilter class."""
        self.all_cols = list(data.columns)
        self.data = data

        # Widgets
        self._add_button = widgets.Button(description="Add filter")
        self._del_button = widgets.Button(description="Delete filter")
        self._upd_button = widgets.Button(description="Update filter")
        self._clear_button = widgets.Button(description="Clear all filters")
        self.apply_button = widgets.Button(description="Apply filter")
        self._col_select = widgets.Dropdown(options=self.all_cols, **(_layout("200px")))
        self._oper_sel = widgets.Dropdown(
            options=self._col_operators(self.current_col), **(_layout("100px"))
        )
        self._not_cb = widgets.Checkbox(
            description="not", value=False, **(_layout("60px", desc_width="initial"))
        )
        self._filter_value = widgets.Textarea(
            description="Filter value", **(_layout("400px"))
        )
        self._curr_filters = widgets.Select(description="Filters", **(_layout("500px")))
        self._oper_label = widgets.Label(" in ")

        self.filters: Dict[str, FilterExpr] = {}

        self._curr_filters.observe(self._select_filter, names="value")
        self._col_select.observe(self._update_operators, names="value")

        self._add_button.on_click(self._add_filter)
        self._upd_button.on_click(self._update_filter)
        self._del_button.on_click(self._del_filter)
        self._clear_button.on_click(self._clear_filters)
        filt_help_lbl = widgets.Label(
            value="Enter multiple values separated by commas. Strings do not need quotes."
        )
        top_row = widgets.VBox(
            [
                filt_help_lbl,
                widgets.HBox(
                    [
                        self._col_select,
                        self._not_cb,
                        self._oper_sel,
                        self._filter_value,
                    ]
                ),
            ]
        )
        mid_row = widgets.HBox(
            [
                self._add_button,
                self._upd_button,
            ]
        )
        curr_filt_lbl = widgets.Label(value="Current filters")
        bottom_row = widgets.VBox(
            [
                curr_filt_lbl,
                widgets.HBox(
                    [
                        self._curr_filters,
                        widgets.VBox([self._del_button, self._clear_button]),
                    ]
                ),
            ],
            **_layout(width="80%", border=True),
        )

        self.layout = widgets.VBox([top_row, mid_row, bottom_row, self.apply_button])

    def display(self):
        """Display in IPython."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()

    def import_filters(self, filters: Dict[str, FilterExpr]):
        """
        Replace the current filters with `filters`.

        Parameters
        ----------
        filters : Dict[str, FilterExpr]
            dict of filter name, FilterExpr
            FilterExpr is a tuple of:
            column [str], inv [bool], operator [str], expr [str]

        """
        self.filters = {
            f_name: FilterExpr(*f_expr) for f_name, f_expr in filters.items()
        }
        self._curr_filters.options = list(filters.keys())

    @property
    def bool_filters(self):
        """Return current set of boolean filters."""
        df_filt = None
        for filt in self.filters.values():
            new_filt = self._make_filter(
                filt.column, filt.operator, filt.expr, filt.inv
            )
            new_filt = new_filt.values if isinstance(new_filt, pd.Series) else new_filt
            df_filt = new_filt if df_filt is None else df_filt & new_filt
        return df_filt if df_filt is not None else self.data.index.isin(self.data.index)

    @property
    def filtered_dataframe(self) -> pd.DataFrame:
        """Return current filtered DataFrame."""
        return self.data[self.bool_filters]

    def _select_filter(self, change):
        filter_name = change["new"]
        if not filter_name:
            return
        (
            self._col_select.value,
            self._not_cb.value,
            self._oper_sel.value,
            self._filter_value.value,
        ) = self.filters[filter_name]

    def _update_operators(self, change):
        del change
        self._oper_sel.options = self._col_operators(self._col_select.value)

    def _add_filter(self, btn):
        del btn
        if self._curr_filter_name in self.filters:
            return
        self.filters[self._curr_filter_name] = FilterExpr(
            column=self._col_select.value,
            inv=self._not_cb.value,
            operator=self._oper_sel.value,
            expr=self._filter_value.value,
        )
        curr_opts = list(self._curr_filters.options)
        curr_opts.append(self._curr_filter_name)
        self._curr_filters.options = curr_opts

    def _update_filter(self, btn):
        selected_filter = self._curr_filters.value
        self._add_filter(btn)
        if selected_filter in self.filters:
            del self.filters[selected_filter]
            self._curr_filters.options = list(self.filters.keys())

    def _del_filter(self, btn):
        del btn
        selected_filter = self._curr_filters.value
        if selected_filter in self.filters:
            del self.filters[selected_filter]
            self._curr_filters.options = list(self.filters.keys())

    @property
    def _curr_filter_name(self):
        not_str = " not " if self._not_cb.value else ""
        return (
            f"{self._col_select.value} {not_str}{self._oper_sel.value}"
            f" '{self._filter_value.value}'"
        )

    def _clear_filters(self, btn):
        del btn
        self.filters.clear()
        self._curr_filters.options = []

    @property
    def current_col(self):
        """Return the currently selected column."""
        return self._col_select.value

    def _col_operators(self, col):
        if pd.api.types.is_string_dtype(self.data[col]):
            return self._OPERATORS["string"]
        return self._OPERATORS["other"]

    def _make_filter(self, col, operator, expr, not_true):
        if not_true:
            return ~self._create_filter(col, operator, expr)
        return self._create_filter(col, operator, expr)

    # pylint: disable=too-many-return-statements
    def _create_filter(self, col: str, operator: str, expr: str) -> pd.Series:
        if operator == "query":
            return pd.Series(self.data.index.isin(self.data.query(expr).index))
        if operator in ("in", "between"):
            return self._filter_in_or_between(col, operator, expr)

        test_expr = self._conv_expr_type(col, expr)
        if operator == "==":
            return self.data[col] == test_expr
        if operator == "contains":
            return self.data[col].str.contains(test_expr)
        if operator == "matches":
            return self.data[col].str.match(test_expr)
        if operator == ">":
            return self.data[col] > test_expr
        if operator == ">=":
            return self.data[col] >= test_expr
        if operator == "<":
            return self.data[col] < test_expr
        if operator == "<=":
            return self.data[col] >= test_expr
        raise TypeError(
            f"Unsupported operator for operator {operator} and column {col}"
        )

    def _filter_in_or_between(self, col: str, operator: str, expr: str) -> pd.Series:
        """Return filter for `in` and `between` operators."""
        test_expr: List[Union[str, int, float]]

        if pd.api.types.is_string_dtype(self.data[col]):
            test_expr = [item.strip("\"' ") for item in expr.split(",")]
        elif pd.api.types.is_numeric_dtype(self.data[col]):
            test_expr = [
                int(item) if "." not in item else float(item)
                for item in expr.split(",")
            ]
        elif pd.api.types.is_datetime64_any_dtype(self.data[col]):
            test_expr = [pd.Timestamp(item.strip()) for item in expr.split(",")]
        else:
            raise TypeError(
                f"Unsupported column type {self.data[col].dtype}",
                f"for operator {operator} and column {col}",
            )
        if operator == "in":
            return self.data[col].isin(test_expr)
        if len(test_expr) != 2:
            raise ValueError(
                f"Must have two operands for expression {expr}",
                f"for operator {operator} and column {col}",
            )
        return self.data[col].between(test_expr[0], test_expr[1], inclusive="both")

    def _conv_expr_type(self, col: str, expr: str):
        """Convert string expression to required type."""
        test_expr: Union[str, int, float]
        if pd.api.types.is_numeric_dtype(self.data[col]):
            test_expr = int(expr) if "." not in expr else float(expr)
        elif pd.api.types.is_datetime64_any_dtype(self.data[col]):
            test_expr = pd.Timestamp(expr.strip())
        elif pd.api.types.is_string_dtype(self.data[col]):
            test_expr = expr.strip("\"' ")
        else:
            raise TypeError(
                f"Unsupported column type {self.data[col].dtype}",
                f"for column {col}",
            )
        return test_expr


def _get_col_width(data, col):
    if data[col].iloc[:10].dropna().empty:
        return 8
    if data[col].dtype == "O":
        return int(data[col].iloc[:10].str.len().mean())
    if pd.api.types.is_datetime64_any_dtype(data[col]):
        return 50
    return 8


def _get_cols_from_df(data):
    """Get list of TableColumn columns from DataFrame."""
    # save the existing column order
    col_order = data.columns
    dt_cols = data.select_dtypes("datetime").columns
    columns = {
        col: TableColumn(field=col, title=col, width=_get_col_width(data, col))
        for col in data.columns
        if col not in dt_cols
    }

    date_fmt = "%F %T"
    dt_columns = {
        col: TableColumn(
            field=col,
            title=col,
            formatter=DateFormatter(format=date_fmt),
            width=_get_col_width(data, col),
        )
        for col in dt_cols
    }
    columns.update(dt_columns)
    return {col: columns[col] for col in col_order}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Tree builder for Windows security and Linux auditd events."""
import attr
import pandas as pd

from ..data.query_defns import ensure_df_datetimes
from .._version import VERSION


__version__ = VERSION
__author__ = "Ian Hellen"


TS_FMT_STRING = "%Y-%m-%d %H:%M:%S.%f"


def extract_process_tree(
    procs: pd.DataFrame,
    schema: "ProcSchema",  # type: ignore  # noqa: F821
    debug: bool = False,
) -> pd.DataFrame:
    """
    Build process trees from the process events.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (Windows 4688 or Linux Auditd)
    schema : Union[ProcSchema, Dict[str, Any]], optional
        The column schema to use, by default None.
        If supplied as a dict it must include definitions for the
        required fields in the ProcSchema class
        If None, then the schema is inferred
    debug : bool
        If True produces extra debugging output,
        by default False

    Returns
    -------
    pd.DataFrame
        Process tree dataframe.

    See Also
    --------
    ProcSchema

    """
    # Clean data
    procs_cln = _clean_proc_data(procs, schema)

    # Merge parent-child
    merged_procs = _merge_parent_by_time(procs_cln, schema)
    if debug:
        _check_merge_status(procs_cln, merged_procs, schema)

    # extract inferred parents
    merged_procs_par = _extract_inferred_parents(merged_procs, schema)
    if debug:
        _check_inferred_parents(merged_procs, merged_procs_par)

    # Create Process and parent Keys
    _assign_proc_key(
        merged_procs_par,
        "proc_key",
        "new_process_lc",
        schema.process_id,
        schema.time_stamp,
    )
    _assign_proc_key(
        merged_procs_par,
        "parent_key",
        "parent_proc_lc",
        schema.parent_id,
        "TimeGenerated_orig_par",
    )
    return merged_procs_par


def _clean_proc_data(
    procs: pd.DataFrame,
    schema: "ProcSchema",  # type: ignore  # noqa: F821
) -> pd.DataFrame:
    """Return cleaned process data."""
    procs = ensure_df_datetimes(procs, columns=schema.time_stamp)
    procs_cln = (
        procs.drop_duplicates().sort_values(schema.time_stamp, ascending=True).copy()
    )

    # Filter out any non-process events
    if schema.event_type_col and schema.event_id_identifier:
        event_type_filter = procs_cln[schema.event_type_col] == schema.event_filter
        procs_cln = procs_cln[event_type_filter]
    # Convert any numeric schema cols to str types
    procs_cln = _num_cols_to_str(procs_cln, schema)

    procs_cln["EffectiveLogonId"] = procs_cln[schema.logon_id]
    # Create effective logon Id for Windows, if the TargetLogonId is not 0x0
    if schema.target_logon_id:
        has_tgt_logonid = (procs_cln[schema.target_logon_id] != "0x0") & (
            procs_cln[schema.target_logon_id].notna()
        )
        procs_cln.loc[has_tgt_logonid, "EffectiveLogonId"] = procs_cln[
            schema.target_logon_id
        ]
    procs_cln["new_process_lc"] = procs_cln[schema.process_name].str.lower()
    if schema.parent_name:
        no_pproc = procs_cln[schema.parent_name] == ""
        procs_cln.loc[no_pproc, schema.parent_name] = "unknown"
        procs_cln["parent_proc_lc"] = procs_cln[schema.parent_name].str.lower()
    procs_cln["source_index"] = procs_cln.index
    return procs_cln


def _num_cols_to_str(
    procs_cln: pd.DataFrame,
    schema: "ProcSchema",  # type: ignore  # noqa: F821
) -> pd.DataFrame:
    """
    Change any numeric columns in our core schema to strings.

    Some columns like PID arrive as float or int types.
    We need to convert the floats to ints (since we want to get rid of the ".0")
    and then convert the int columns to str. We need to do this since
    we build parent/child keys by concatenating some of these value
    into a single string.
    """
    # Change float/int cols in our core schema to force int
    schema_cols = [
        col for col in attr.asdict(schema).values() if col and col in procs_cln.columns
    ]
    force_int_cols = {
        col: "int"
        for col, col_type in procs_cln[schema_cols].dtypes.to_dict().items()
        if pd.api.types.is_float_dtype(col_type)
    }
    procs_cln = procs_cln.astype(force_int_cols)
    # then change any int types in to string types
    # note: we need the prev
    int_to_str_cols = {
        col: "str"
        for col, col_type in procs_cln[schema_cols].dtypes.to_dict().items()
        if pd.api.types.is_integer_dtype(col_type)
    }
    return procs_cln.astype(int_to_str_cols)


def _merge_parent_by_time(
    procs: pd.DataFrame,
    schema: "ProcSchema",  # type: ignore  # noqa: F821
) -> pd.DataFrame:
    """Merge procs with parents using merge_asof."""
    parent_procs = (
        procs[
            [
                schema.process_id,
                "EffectiveLogonId",
                "new_process_lc",
                "source_index",
                schema.parent_id,
                schema.time_stamp,
                schema.process_name,
            ]
        ]
        .assign(TimeGenerated_orig_par=procs[schema.time_stamp])
        .sort_values(schema.time_stamp, ascending=True)
    )
    # if we have a parent name (Windows) - use that as part of the
    # match
    if schema.parent_name:
        par_join_cols = [schema.process_id, "new_process_lc"]
        child_join_cols = [schema.parent_id, "parent_proc_lc"]
    else:
        par_join_cols = [schema.process_id]
        child_join_cols = [schema.parent_id]
    # merge_asof merges on the "by" fields and then the closest time
    # match in the time_stamp field. The default is to look backwards
    # for a match on the right of the join (parent) that is a time earlier
    # than the corresponding row on the left.
    return pd.merge_asof(
        left=procs.sort_values(schema.time_stamp, ascending=True),
        right=parent_procs,
        on=schema.time_stamp,
        left_by=child_join_cols,
        right_by=par_join_cols,
        suffixes=("", "_par"),
    )


def _extract_inferred_parents(
    merged_procs: pd.DataFrame, schema: "ProcSchema"  # type: ignore  # noqa: F821
) -> pd.DataFrame:
    """Find any inferred parents and creates rows for them."""
    tz_aware = merged_procs.iloc[0][schema.time_stamp].tz
    time_zero = pd.Timestamp(0) if tz_aware is None else pd.Timestamp(0, tz=0)

    # Fill in missing values for root processes
    root_procs_crit = merged_procs["source_index_par"].isna()
    merged_procs.loc[root_procs_crit, "NewProcessId_par"] = merged_procs[
        schema.parent_id
    ]
    if schema.parent_name:
        merged_procs.loc[root_procs_crit, "new_process_lc_par"] = merged_procs[
            "parent_proc_lc"
        ]
    else:
        merged_procs.loc[root_procs_crit, "new_process_lc_par"] = "unknown"
        merged_procs.loc[root_procs_crit, f"{schema.process_name}_par"] = "unknown"
        # If the schema doesn't have a ParentProcessName/parent_proc_lc - copy this value
        # from the merged data for ALL processes
        merged_procs["ParentProcessName"] = merged_procs[f"{schema.process_name}_par"]
        merged_procs["parent_proc_lc"] = merged_procs["new_process_lc_par"]
    merged_procs.loc[root_procs_crit, "EffectiveLogonId_par"] = merged_procs[
        schema.logon_id
    ]
    merged_procs.loc[root_procs_crit, "TimeGenerated_orig_par"] = time_zero

    # Extract synthentic rows for the parents of root processes
    inferred_parents = (
        merged_procs[root_procs_crit][
            [
                schema.event_id_column,
                schema.host_name_column,
                schema.parent_id,
                "EffectiveLogonId_par",
                "ParentProcessName",
                "parent_proc_lc",
            ]
        ]
        .rename(
            columns={
                schema.parent_id: schema.process_id,
                "ParentProcessName": schema.process_name,
                "parent_proc_lc": "new_process_lc",
                "EffectiveLogonId_par": schema.logon_id,
            }
        )
        .assign(TimeGenerated=time_zero, EffectiveLogonId=merged_procs[schema.logon_id])
        .drop_duplicates()
    )

    return pd.concat(
        [merged_procs, inferred_parents], ignore_index=True, axis=0, sort=False
    )


def _assign_proc_key(
    proc_data: pd.DataFrame,
    key_name: str,
    proc_name_col: str,
    proc_id_col: str,
    timestamp_col: str,
):
    """Create process and parent keys for unambiguous par-child relation."""
    proc_data[key_name] = (
        proc_data[proc_name_col]
        + "|"
        + proc_data[proc_id_col].astype(str)
        + "|"
        + proc_data[timestamp_col].dt.round("10us").dt.strftime(TS_FMT_STRING)
    )


# Diagnostic/debug functions
def _check_merge_status(procs, merged_procs, schema):
    """Diagnostic for _merge_parent_by_time."""
    orig_cols = [col for col in merged_procs.columns if not col.endswith("_par")]
    rows_with_dups2 = (
        merged_procs.dropna()
        .groupby(orig_cols)
        .count()
        .reset_index()
        .query(f"{schema.process_id}_par > 1")["source_index"]
    )

    # Check status
    print("Original # procs", len(procs))
    print("Merged # procs", len(merged_procs))
    print("Merged # procs - dropna", len(merged_procs.dropna()))

    print(
        "Unique merged_procs index in merge", len(merged_procs["source_index"].unique())
    )
    print("These two should add up to top line")
    row_dups = len(rows_with_dups2)
    print("Rows with dups", row_dups)
    row_nodups = len(merged_procs[~merged_procs["source_index"].isin(rows_with_dups2)])
    print("Rows with no dups", row_nodups)
    print(row_dups, "+", row_nodups, "=", row_dups + row_nodups)


def _check_inferred_parents(procs, procs_par):
    """Diagnostic for _extract_inferred_parents."""
    print(
        "original:",
        len(procs),
        "inferred_parents",
        len(procs_par) - len(procs),
        "combined",
        len(procs_par),
    )


def _check_proc_keys(merged_procs_par, schema):
    """Diagnostic for _assign_proc_keys."""
    crit1 = merged_procs_par["TimeGenerated_orig_par"].isin(
        merged_procs_par[schema.time_stamp]
    )
    crit2 = merged_procs_par["EffectiveLogonId"].isin(merged_procs_par[schema.logon_id])
    c2a = None
    if schema.target_logon_id:
        c2a = merged_procs_par["EffectiveLogonId"].isin(
            merged_procs_par[schema.target_logon_id]
        )
    crit3 = merged_procs_par["parent_proc_lc"].isin(merged_procs_par["new_process_lc"])
    crit4 = merged_procs_par[schema.process_id].isin(merged_procs_par[schema.parent_id])
    crit5 = merged_procs_par["parent_key"].isin(merged_procs_par.index)
    crit6 = merged_procs_par["parent_key"].isna()
    print("has parent time", len(merged_procs_par[crit1]))
    print("effectivelogonId in subjectlogonId", len(merged_procs_par[crit2]))
    if schema.target_logon_id and c2a is not None:
        print("effectivelogonId in targetlogonId", len(merged_procs_par[c2a]))
    print("parent_proc_lc in procs", len(merged_procs_par[crit3]))
    print("ProcessId in ParentProcessId", len(merged_procs_par[crit4]))
    print("Parent_key in proc_key", len(merged_procs_par[crit5]))
    print("Parent_key not in proc_key", len(merged_procs_par[~crit5]))
    print("Parent_key is NA", len(merged_procs_par[crit6]))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Host Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .entity_enums import OSFamily

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class Host(Entity):
    """
    Host Entity class.

    Attributes
    ----------
    DnsDomain : str
        Host DnsDomain
    NTDomain : str
        Host NTDomain
    HostName : str
        Host HostName
    NetBiosName : str
        Host NetBiosName
    AzureID : str
        Host AzureID
    OMSAgentID : str
        Host OMSAgentID
    OSFamily : str
        Host OSFamily
    OSVersion : str
        Host OSVersion
    IsDomainJoined : bool
        Host IsDomainJoined

    """

    ID_PROPERTIES = ["fqdn", "AzureID", "OMSAgentID"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.DnsDomain: Optional[str] = None
        self.NTDomain: Optional[str] = None
        self.HostName: Optional[str] = None
        self.NetBiosName: Optional[str] = None
        self.AzureID: Optional[str] = None
        self.OMSAgentID: Optional[str] = None
        self.OSFamily: OSFamily = OSFamily.Windows
        self.OSVersion: Optional[str] = None
        self.IsDomainJoined: bool = False

        super().__init__(src_entity=src_entity, **kwargs)
        self._computer = None
        if src_event is not None:
            self._create_from_event(src_event)

    @property
    def computer(self) -> Optional[str]:
        """Return computer from source event."""
        return self._computer if self._computer is not None else self.fqdn

    @property
    def fqdn(self) -> Optional[str]:
        """Construct FQDN from host + dns."""
        if self.DnsDomain:
            return f"{self.HostName}.{self.DnsDomain}"
        return self.HostName

    @property
    def FullName(self) -> Optional[str]:  # noqa: N802
        """Return the full name of the host - either FQDN or Netbiosname."""
        if self.DnsDomain:
            return f"{self.HostName or self.NetBiosName}.{self.DnsDomain}"
        if self.NTDomain:
            return f"{self.HostName or self.NetBiosName}.{self.NTDomain}"
        return self.HostName or self.NetBiosName

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.fqdn} ({self.OSFamily})"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.HostName or self.__class__.__name__

    def _create_from_event(self, src_event):
        if "Computer" in src_event:
            self._computer = src_event["Computer"]
            if "." in src_event["Computer"]:
                self.HostName = src_event["Computer"].split(".", 1)[0]
                self.DnsDomain = src_event["Computer"].split(".", 1)[1]
            else:
                self.HostName = src_event["Computer"]
        elif "HostName" in src_event:
            self.HostName = src_event["HostName"]
            if "DnsDomain" in src_event:
                self.DnsDomain = src_event["DnsDomain"]
        self.NetBiosName = self.HostName

    _entity_schema = {
        # DnsDomain (type System.String)
        "DnsDomain": None,
        # NTDomain (type System.String)
        "NTDomain": None,
        # HostName (type System.String)
        "HostName": None,
        # NetBiosName (type System.String)
        "NetBiosName": None,
        # AzureID (type System.String)
        "AzureID": None,
        # OMSAgentID (type System.String)
        "OMSAgentID": None,
        # OSFamily (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.OSFamily])
        "OSFamily": "OSFamily",
        # IsDomainJoined (type System.Nullable`1[System.Boolean])
        "IsDomainJoined": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure KeyVault pre-authentication."""
import logging
import sys
from collections import namedtuple
from datetime import datetime
from enum import Enum
from typing import List, Optional, Tuple

from azure.common.exceptions import CloudError
from azure.common.credentials import get_cli_profile
from azure.identity import (
    AzureCliCredential,
    ChainedTokenCredential,
    EnvironmentCredential,
    InteractiveBrowserCredential,
    ManagedIdentityCredential,
)
from dateutil import parser
from msrestazure import azure_cloud

from .._version import VERSION
from .cred_wrapper import CredentialWrapper
from .exceptions import MsticpyAzureConnectionError
from ..common import pkg_config as config
from .cloud_mappings import (
    get_all_endpoints,
    get_all_suffixes,
    CLOUD_ALIASES,
    CLOUD_MAPPING,
)

__version__ = VERSION
__author__ = "Pete Bryan"


AzCredentials = namedtuple("AzCredentials", ["legacy", "modern"])


def default_auth_methods() -> List[str]:
    """Get the default (all) authentication options."""
    try:
        az_settings = config.get_config("Azure")
        if az_settings and "auth_methods" in az_settings:
            return az_settings["auth_methods"]
    except KeyError:
        pass  # no Azure section in config
    return ["cli", "msi", "interactive"]


class AzureCloudConfig:
    """Azure Cloud configuration."""

    def __init__(self, cloud: str = None):
        """
        Initialize AzureCloudConfig from `cloud` or configuration.

        Parameters
        ----------
        cloud : str, optional
            The cloud to retrieve configuration for. If not supplied,
            the cloud ID is read from configuration. If this is not available,
            it defaults to 'global'.

        """
        if cloud:
            self.cloud = cloud
        else:
            self.cloud = "global"
            try:
                az_settings = config.get_config("Azure")
                if az_settings and "cloud" in az_settings:
                    self.cloud = az_settings["cloud"]
            except KeyError:
                pass  # no Azure section in config
        self.auth_methods = default_auth_methods()
        try:
            self.auth_methods = config.get_config("Azure").get(
                "auth_methods", default_auth_methods()
            )
        except KeyError:
            pass  # no Azure section in config

    @property
    def cloud_names(self) -> List[str]:
        """Return a list of current cloud names."""
        return list(CLOUD_MAPPING.keys())

    @staticmethod
    def resolve_cloud_alias(alias) -> Optional[str]:
        """Return match of cloud alias or name."""
        alias_cf = alias.casefold()
        aliases = {alias.casefold(): cloud for alias, cloud in CLOUD_ALIASES.items()}
        if alias_cf in aliases:
            return aliases[alias_cf]
        if alias_cf in aliases.values():
            return alias_cf
        return None

    @property
    def endpoints(self) -> azure_cloud.CloudEndpoints:
        """
        Get a list of all the endpoints for an Azure cloud.

        Returns
        -------
        dict
            A dictionary of endpoints for the cloud.

        Raises
        ------
        MsticpyAzureConfigError
            If the cloud name is not valid.

        """
        return get_all_endpoints(self.cloud)

    @property
    def suffixes(self) -> azure_cloud.CloudSuffixes:
        """
        Get a list of all the suffixes for an Azure cloud.

        Returns
        -------
        dict
            A dictionary of suffixes for the cloud.

        Raises
        ------
        MsticpyAzureConfigError
            If the cloud name is not valid.

        """
        return get_all_suffixes(self.cloud)

    @property
    def token_uri(self) -> str:
        """Return the resource manager token URI."""
        return f"{self.endpoints.resource_manager}.default"


def _az_connect_core(
    auth_methods: List[str] = None, cloud: str = None, silent: bool = False, **kwargs
) -> AzCredentials:
    """
    Authenticate using multiple authentication sources.

    Parameters
    ----------
    auth_methods : List[str], optional
        List of authentication methods to try
        Possible options are:
        - "env" - to get authentication details from environment variables
        - "cli" - to use Azure CLI authentication details
        - "msi" - to user Managed Service Identity details
        - "interactive" - to prompt for interactive login
        If not set, it will use the value defined in msticpyconfig.yaml.
        If this is not set, the default is ["env", "cli", "msi", "interactive"]
    cloud : str, optional
        What Azure cloud to connect to.
        By default it will attempt to use the cloud setting from config file.
        If this is not set it will default to Azure Public Cloud
    silent : bool, optional
        Whether to display any output during auth process. Default is False.

    Returns
    -------
    AzCredentials
                Named tuple of:
        - legacy (ADAL) credentials
        - modern (MSAL) credentials

    Raises
    ------
    MsticpyAzureConnectionError
        If invalid auth options are presented.
    CloudError
        If chained token credential creation fails.

    Notes
    -----
    The function tries to obtain credentials from the following
    sources:
    - Azure Auth Environment variables
    - Azure CLI (if an active session is logged on)
    - Managed Service Identity
    - Interactive browser logon
    If the authentication is successful both ADAL (legacy) and
    MSAL (modern) credential types are returned.

    """
    # Create the auth methods with the specified cloud region
    cloud = cloud or kwargs.pop("region", AzureCloudConfig().cloud)
    auth_options = _create_auth_options(cloud)
    if not auth_methods:
        auth_methods = default_auth_methods()
    try:
        auths = [auth_options[meth] for meth in auth_methods]
    except KeyError as err:
        raise MsticpyAzureConnectionError(
            "Unknown authentication option, valid options are; env, cli, msi, interactive"
        ) from err

    # Filter and replace error message when credentials not found
    handler = logging.StreamHandler(sys.stdout)
    if silent:
        handler.addFilter(_filter_all_warnings)
    else:
        handler.addFilter(_filter_credential_warning)
    logging.basicConfig(level=logging.WARNING, handlers=[handler])

    # Create credentials and connect to the subscription client to validate
    creds = ChainedTokenCredential(*auths)  # type: ignore
    legacy_creds = CredentialWrapper(
        creds, resource_id=AzureCloudConfig(cloud).token_uri
    )
    if not creds:
        raise CloudError("Could not obtain credentials.")

    return AzCredentials(legacy_creds, creds)


class _AzCachedConnect:
    """Singleton class caching Azure credentials."""

    _instance = None

    def __new__(cls):
        """Override new to check and return existing instance."""
        if cls._instance is None:
            cls._instance = super(_AzCachedConnect, cls).__new__(cls)
            cls.connect.__doc__ = _az_connect_core.__doc__
        return cls._instance

    def __init__(self):
        """Initialize the class."""
        self.az_credentials: Optional[AzCredentials] = None
        self.cred_cloud: str = self.current_cloud

    @property
    def current_cloud(self) -> str:
        """Return current cloud."""
        return AzureCloudConfig().cloud

    def connect(self, *args, **kwargs):
        """Call az_connect_core if token is not present or expired."""
        if self.az_credentials is None:
            self.az_credentials = _az_connect_core(*args, **kwargs)
            return self.az_credentials
        # Check expiry
        if (
            datetime.utcfromtimestamp(
                self.az_credentials.modern.get_token(
                    AzureCloudConfig().token_uri
                ).expires_on
            )
            <= datetime.utcnow()
        ):
            self.az_credentials = _az_connect_core(*args, **kwargs)
        # Check changed cloud
        if self.cred_cloud != kwargs.get(
            "cloud", kwargs.get("region", self.current_cloud)
        ):
            self.az_credentials = _az_connect_core(*args, **kwargs)
        return self.az_credentials


# externally callable function using the class above
_AZ_CACHED_CONNECT = _AzCachedConnect()
az_connect_core = _AZ_CACHED_CONNECT.connect


def only_interactive_cred(chained_cred: ChainedTokenCredential):
    """Return True if only interactivebrowser credentials available."""
    return len(chained_cred.credentials) == 1 and isinstance(
        chained_cred.credentials[0], InteractiveBrowserCredential
    )


def _filter_credential_warning(record) -> bool:
    """Rewrite out credential not found message."""
    if (
        not record.name.startswith("azure.identity")
        or record.levelno != logging.WARNING
    ):
        return True
    message = record.getMessage()
    if ".get_token" in message:
        if message.startswith("EnvironmentCredential"):
            print("Attempting to sign-in with environment variable credentials...")
        if message.startswith("AzureCliCredential"):
            print("Attempting to sign-in with Azure CLI credentials...")
        if message.startswith("ManagedIdentityCredential"):
            print("Attempting to sign-in with Managed Instance credentials...")
            print("Falling back to interactive logon.")
    return not message


def _filter_all_warnings(record) -> bool:
    """Filter out credential error messages."""
    if record.name.startswith("azure.identity") and record.levelno == logging.WARNING:
        message = record.getMessage()
        if ".get_token" in message:
            return not message
    return True


def _create_auth_options(cloud: str = None) -> dict:
    """Create auth options dict with correct cloud set."""
    az_config = AzureCloudConfig(cloud)

    aad_uri = az_config.endpoints.active_directory  # type: ignore

    return {
        "env": EnvironmentCredential(),
        "cli": AzureCliCredential(),
        "msi": ManagedIdentityCredential(),
        "interactive": InteractiveBrowserCredential(authority=aad_uri),
    }


class AzureCliStatus(Enum):
    """Enumeration for _check_cli_credentials return values."""

    CLI_OK = 0
    CLI_NOT_INSTALLED = 1
    CLI_NEEDS_SIGN_IN = 2
    CLI_TOKEN_EXPIRED = 3
    CLI_UNKNOWN_ERROR = 4


def check_cli_credentials() -> Tuple[AzureCliStatus, Optional[str]]:
    """Check to see if there is a CLI session with a valid AAD token."""
    try:
        cli_profile = get_cli_profile()
        raw_token = cli_profile.get_raw_token()
        bearer_token = None
        if (
            isinstance(raw_token, tuple)
            and len(raw_token) == 3
            and len(raw_token[0]) == 3
        ):
            bearer_token = raw_token[0][2]
            if (
                parser.parse(bearer_token.get("expiresOn", datetime.min))
                < datetime.now()
            ):
                raise ValueError("AADSTS70043: The refresh token has expired")

        return AzureCliStatus.CLI_OK, "Azure CLI credentials available."
    except ImportError:
        # Azure CLI not installed
        return AzureCliStatus.CLI_NOT_INSTALLED, None
    except Exception as ex:  # pylint: disable=broad-except
        if "AADSTS70043: The refresh token has expired" in str(ex):
            message = (
                "Azure CLI was detected but the token has expired. "
                "For Azure CLI single sign-on, please sign in using '!az login'."
            )
            return AzureCliStatus.CLI_TOKEN_EXPIRED, message
        if "Please run 'az login' to setup account" in str(ex):
            message = (
                "Azure CLI was detected but no token is available. "
                "For Azure CLI single sign-on, please sign in using '!az login'."
            )
            return AzureCliStatus.CLI_NEEDS_SIGN_IN, message
        return AzureCliStatus.CLI_UNKNOWN_ERROR, None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Outlier detection class. TODO **Preliminary**.

Similar to the eventcluster module but a little bit more experimental
(read 'less tested'). It uses SkLearn Isolation Forest to identify
outlier events in a single data set or using one data set as training
data and another on which to predict outliers.

"""

import math
from typing import List, Tuple

import pandas as pd
import numpy as np

from ..common.exceptions import MsticpyImportExtraError
from .._version import VERSION

try:
    import matplotlib.pyplot as plt
    from sklearn.ensemble import IsolationForest
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without Sklearn and matplotlib installed",
        title="Error importing Scikit Learn and matplotlib",
        extra="ml",
    ) from imp_err

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name
def identify_outliers(
    x: np.ndarray, x_predict: np.ndarray, contamination: float = 0.05
) -> Tuple[IsolationForest, np.ndarray, np.ndarray]:
    """
    Identify outlier items using SkLearn IsolationForest.

    Arguments
    ---------
    x : np.ndarray
            Input data
    x_predict : np.ndarray
        Model
    contamination : float
        Percentage contamination (default: {0.05})

    Returns
    -------
    Tuple[IsolationForest, np.ndarray, np.ndarray]
        IsolationForest model, X_Outliers,
        y_pred_outliers

    """
    # pylint: disable=no-member
    rng = np.random.RandomState(42)

    # fit the model
    rows, cols = x.shape
    max_samples = min(100, cols)
    max_features = math.floor(math.sqrt(rows))
    clf = IsolationForest(
        max_samples=max_samples,
        max_features=max_features,
        random_state=rng,
        contamination=contamination,
    )

    # fit and train the model
    clf.fit(x)
    clf.predict(x)

    y_pred_outliers = clf.predict(x_predict)

    x_outliers = x_predict[y_pred_outliers == -1]
    return clf, x_outliers, y_pred_outliers


# pylint: disable=too-many-arguments, too-many-locals
def plot_outlier_results(
    clf: IsolationForest,
    x: np.ndarray,
    x_predict: np.ndarray,
    x_outliers: np.ndarray,
    feature_columns: List[int],
    plt_title: str,
):
    """
    Plot Isolation Forest results.

    Parameters
    ----------
    clf : IsolationForest
        Isolation Forest model
    x : np.ndarray
        Input data
    x_predict : np.ndarray
        Prediction
    x_outliers : np.ndarray
        Set of outliers
    feature_columns : List[int]
        list of feature columns to display
    plt_title : str
        Plot title

    """
    # plot the line, the samples, and the nearest vectors to the plane
    x_max_x = x[:, 0].max() + (x[:, 0].max() / 10)
    x_min_x = -x[:, 0].max() / 10
    x_max_y = x[:, 1].max() + (x[:, 1].max() / 10)
    x_min_y = -x[:, 1].max() / 10
    xx, yy = np.meshgrid(
        np.linspace(x_min_x, x_max_x, 100), np.linspace(x_min_y, x_max_y, 100)
    )
    z = clf.decision_function(np.c_[xx.ravel(), yy.ravel()])
    z = z.reshape(xx.shape)

    plt.rcParams["figure.figsize"] = (20, 10)

    plt.title(plt_title)
    # pylint: disable=no-member
    plt.contourf(xx, yy, z, cmap=plt.cm.Blues_r)

    b1 = plt.scatter(x[:, 0], x[:, 1], c="white", s=20, edgecolor="k")
    b2 = plt.scatter(x_predict[:, 0], x_predict[:, 1], c="green", s=40, edgecolor="k")
    c = plt.scatter(
        x_outliers[:, 0], x_outliers[:, 1], c="red", marker="x", s=200, edgecolor="k"
    )
    plt.axis("tight")

    xp_max_x = x_predict[:, 0].max() + (x_predict[:, 0].max() / 10)
    xp_min_x = -x_predict[:, 0].max() / 10
    xp_max_y = x_predict[:, 1].max() + (x_predict[:, 1].max() / 10)
    xp_min_y = -x_predict[:, 1].max() / 10

    plt.xlim((xp_min_x, xp_max_x))
    plt.ylim((xp_min_y, xp_max_y))
    plt.xlabel(feature_columns[0])
    plt.ylabel(feature_columns[1])

    plt.legend(
        [b1, b2, c],
        [
            "training observations",
            "new regular observations",
            "new abnormal observations",
        ],
        loc="upper right",
    )
    plt.show()


def remove_common_items(data: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """
    Remove rows from input DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        Input dataframe
    columns : List[str]
        Column list to filter

    Returns
    -------
    pd.DataFrame
        Filtered DataFrame

    """
    filtered_df = data
    # pylint: disable=cell-var-from-loop
    for col in columns:
        filtered_df = filtered_df.filter(
            lambda x: (x[col].std() == 0 and x[col].count() > 10)
        )

    return filtered_df



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Keyvault client - adapted from Bluehound code."""
import base64
import json
from typing import Any, List

from azure.core.exceptions import ResourceNotFoundError
from azure.keyvault.secrets import KeyVaultSecret, SecretClient

from azure.mgmt.keyvault import KeyVaultManagementClient
from azure.mgmt.keyvault.models import (
    AccessPolicyEntry,
    CertificatePermissions,
    KeyPermissions,
    Permissions,
    SecretPermissions,
    Sku,
    Vault,
    VaultCreateOrUpdateParameters,
    VaultProperties,
)

from msrestazure.azure_exceptions import CloudError

from .._version import VERSION
from .azure_auth_core import az_connect_core
from .exceptions import MsticpyKeyVaultConfigError, MsticpyKeyVaultMissingSecretError
from .keyvault_settings import KeyVaultSettings
from .utility import export

__version__ = VERSION
__author__ = "Matt Richard, Ian Hellen"


@export
class BHKeyVaultClient:
    """Core KeyVault client."""

    _KEYRING_NAME = "keyvault"

    def __init__(
        self,
        tenant_id: str = None,
        vault_uri: str = None,
        vault_name: str = None,
        settings: KeyVaultSettings = None,
        **kwargs,
    ):
        """
        Initialize the BHKeyVault client.

        Parameters
        ----------
        tenant_id : str
            The tenant ID of the service
        vault_uri : str, optional
            The full URI of the keyvault, by default None
        vault_name : str, optional
            The name of the keyvault in the public cloud, by default None
        auth_methods : List[str]
            The authentication methods to use for Key Vault auth
            Possible values are:
            - "env" - to get authentication details from environment varibales
            - "cli" - to use Azure CLI authentication details
            - "msi" - to user Managed Service Indenity details
            - "interactive" - to prompt for interactive login
        authn_type : str, optional
            [deprecated - use auth_methods]
            Authentication mode, by default 'interactive'
            Supported options are:
            - 'device' for device code authentication
            - 'interactive' for interactive browser authentication
        authority : str, optional
            The AAD authority - one of 'global', 'usgov', 'de' or 'chi'
        authority_uri : str, optional
            The AAD authority URI - overrides `authority`
        settings : KeyVaultSettings
            An instance of KeyVaultSettings containing KV parameters.
        debug : bool, optional
            [description], by default False

        Raises
        ------
        KeyVaultMissingVaultException
            No Vault name or URI supplied.

        Notes
        -----
        The parameter values can also be obtained from the
        KeyVault section of msticpyconfig.yaml.


        """
        self.debug = kwargs.pop("debug", False)
        self.settings: KeyVaultSettings = settings or KeyVaultSettings()

        self.tenant_id = tenant_id or self.settings.get("tenantid")
        if not self.tenant_id:
            raise MsticpyKeyVaultConfigError(
                "Could not get TenantId from function parameters or configuration.",
                "Please add this to the KeyVault section of msticpyconfig.yaml",
                title="missing tenant ID value.",
            )
        self.authn_type = kwargs.pop(
            "authn_type", self.settings.get("authntype", "interactive")
        )
        self.auth_methods = kwargs.pop(
            "auth_methods", self.settings.get("auth_methods", ["interactive"])
        )

        # for authority and authority_uri, any parameters take priority
        # and fall back on settings if not specified.
        if "authority" in kwargs:
            self.settings["authority"] = kwargs.pop("authority")
        self.authority_uri = self.settings.get_tenant_authority_host(
            authority_uri=kwargs.get("authority_uri"), tenant=self.tenant_id
        )

        if not vault_uri and not vault_name:
            if "vaultname" in self.settings:
                vault_name = self.settings["vaultname"]
            else:
                raise MsticpyKeyVaultConfigError(
                    "Check that you have specified the right value for VaultName"
                    + " in your configuration",
                    title="Key Vault vault name not found.",
                )
        if vault_uri:
            self.vault_uri = vault_uri
        else:
            vault_uri = self.settings.keyvault_uri
            if vault_uri:
                self.vault_uri = vault_uri.format(vault=vault_name)
            else:
                cloud = self.settings.cloud
                raise MsticpyKeyVaultConfigError(
                    f"Could not determine keyvault URI for national cloud {cloud}.",
                    "Please verify that you have the correct national cloud"
                    + "specified in the KeyVault section of msticpyconfig.yaml",
                    title="no Key Vault URI for national cloud",
                )
        if self.debug:
            print(f"Using Vault URI {self.vault_uri}")

        self.kv_client = self._get_secret_client()

    def _get_secret_client(self):
        credentials = az_connect_core(auth_methods=self.auth_methods)

        # Create a secret client
        return SecretClient(self.vault_uri, credentials.modern)

    @property
    def secrets(self):
        """Return the list of secret names from the vault."""
        return [x.id for x in self.kv_client.list_properties_of_secrets()]

    def get_secret(self, secret_name: str) -> Any:
        """
        Retrieve a secret from the Vault.

        Parameters
        ----------
        secret_name : str
            Name of the secret

        Returns
        -------
        Any
            The secret value

        Raises
        ------
        KeyVaultMissingSecretException
            Secret not found in the Vault.

        """
        if "/" in secret_name:
            # If we're passed the full URL to the secret - extract just the
            # name
            secret_name = secret_name.rsplit("/", maxsplit=1)[-1]
        try:
            secret_bundle = self.kv_client.get_secret(name=secret_name)
        except ResourceNotFoundError as err:
            if self.debug:
                print(f"Secret: '{secret_name}' missing from vault: {self.vault_uri}")
            raise MsticpyKeyVaultMissingSecretError(
                f"Secret name {secret_name} could not be found in {self.vault_uri}",
                f"Provider returned: {err}",
                title=f"secret {secret_name} not found.",
            ) from err
        if secret_bundle.value is None or not secret_bundle.value:
            if self.debug:
                print(f"Secret: '{secret_name}' was empty in vault {self.vault_uri}")
            raise MsticpyKeyVaultMissingSecretError(
                f"Secret name {secret_name} in {self.vault_uri}",
                "has blank or null value.",
                title=f"secret {secret_name} empty.",
            )
        return secret_bundle.value

    def set_secret(self, secret_name: str, value: Any) -> KeyVaultSecret:
        """
        Set a secret in the Vault.

        Parameters
        ----------
        secret_name : str
            Name of the secret
        value: Any
            Secret value

        Returns
        -------
        KeyVaultSecret
            The secrets bundle for the secret

        """
        if self.debug:
            print(f"Storing {secret_name} in {self.vault_uri}")
        return self.kv_client.set_secret(name=secret_name, value=value)


# pylint: disable=too-many-instance-attributes
@export
class BHKeyVaultMgmtClient:
    """Core KeyVault Management client."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        tenant_id: str = None,
        subscription_id: str = None,
        resource_group: str = None,
        azure_region: str = None,
        settings: KeyVaultSettings = None,
        **kwargs,
    ):
        """
        Initialize BH KeyVault Management Client.

        Parameters
        ----------
        tenant_id : str, Optional
            Tenant ID
        subscription_id : str, Optional
            Subscription ID
        resource_group : str, Optional
            Resource Group name
        azure_region : str, Optional
            Azure region - needed to create a new vault.
            By default, None
        settings : KeyVaultSettings
            An instance of KeyVaultSettings containing KV parameters.
        mgmt_uri : str, Optional
            The URI for Azure management endpoints.

        Notes
        -----
        The parameter values can also be obtained from the
        KeyVault section of msticpyconfig.yaml.

        """
        self.debug = kwargs.pop("debug", False)
        self.settings: KeyVaultSettings = settings or KeyVaultSettings()
        self.tenant_id = tenant_id or self.settings.get("tenantid")
        if not self.tenant_id:
            raise MsticpyKeyVaultConfigError(
                "Could not get TenantId from function parameters or configuration.",
                "Please add this to the KeyVault section of msticpyconfig.yaml",
                title="missing tenant ID value.",
            )
        self.subscription_id = subscription_id or self.settings.get("subscriptionid")
        if not self.subscription_id:
            raise MsticpyKeyVaultConfigError(
                "Could not get SubscriptionId from function parameters or configuration.",
                "Please add this to the KeyVault section of msticpyconfig.yaml",
                title="missing SubscriptionId value.",
            )
        self._client_uri = kwargs.pop("mgmt_uri", None) or self.settings.mgmt_uri
        if not self._client_uri:
            cloud = self.settings.cloud
            raise MsticpyKeyVaultConfigError(
                f"Could not obtain an azure management URI for national cloud {cloud}.",
                "Please verify that you have the correct national cloud"
                + "specified in the KeyVault section of msticpyconfig.yaml",
                title="no Azure Management URI for national cloud",
            )

        self.auth_client = az_connect_core()
        self.resource_group = resource_group or self.settings.get("resourcegroup")
        self.azure_region = azure_region or self.settings.get("azureregion")

    # pylint: enable=too-many-arguments

    def list_vaults(self) -> List[str]:
        """
        Return a list of vaults for the subscription.

        Returns
        -------
        List[str]
            Vault names

        """
        mgmt = KeyVaultManagementClient(self.auth_client.legacy, self.subscription_id)
        return [v.name for v in mgmt.vaults.list()]

    def get_vault_uri(self, vault_name: str) -> str:
        """
        Return the URI for a vault name.

        Parameters
        ----------
        vault_name : str
            The Vault name.

        Returns
        -------
        str
            Vault URI.

        """
        mgmt = KeyVaultManagementClient(self.auth_client.legacy, self.subscription_id)
        try:
            vault = mgmt.vaults.get(self.resource_group, vault_name)
        except (CloudError, ResourceNotFoundError) as cloud_err:
            raise MsticpyKeyVaultConfigError(
                "Check that you have specified the right value for VaultName"
                + " in your configuration",
                f"Error returned from provider was {cloud_err}",
                title=f"Key Vault vault '{vault_name}' not found.",
            ) from cloud_err
        return vault.properties.vault_uri

    def create_vault(self, vault_name: str) -> Vault:
        """
        Create new or update existing vault.

        Parameters
        ----------
        vault_name : str
            Name of the Vault

        Returns
        -------
        Vault
            The Vault object.

        """
        if not self.azure_region:
            raise MsticpyKeyVaultConfigError(
                "Could not get Azure region in which to create the vault.",
                "Please add AzureRegion to the KeyVault section of msticpyconfig.yaml",
                title="missing AzureRegion value.",
            )
        parameters = self._get_params()
        if not self.resource_group:
            raise MsticpyKeyVaultConfigError(
                "Could not get Azure resource group in which to create the vault.",
                "Please add ResourceGroup to the KeyVault section of msticpyconfig.yaml",
                title="missing ResourceGroup value.",
            )
        mgmt = KeyVaultManagementClient(self.auth_client.legacy, self.subscription_id)
        return mgmt.vaults.create_or_update(
            self.resource_group, vault_name, parameters
        ).result()

    def _get_params(self):
        """Build the vault parameters block."""
        oid = _user_oid(self.auth_client.legacy.token)
        sec_perms_all = [perm.value for perm in SecretPermissions]
        key_perms_all = [perm.value for perm in KeyPermissions]
        cert_perms_all = [perm.value for perm in CertificatePermissions]
        permissions = Permissions()
        permissions.keys = key_perms_all
        permissions.secrets = sec_perms_all
        permissions.certificates = cert_perms_all

        policy = AccessPolicyEntry(
            tenant_id=self.tenant_id, object_id=oid, permissions=permissions
        )

        properties = VaultProperties(
            tenant_id=self.tenant_id,
            sku=Sku(name="standard", family="A"),
            access_policies=[policy],
        )
        parameters = VaultCreateOrUpdateParameters(
            location=self.azure_region, properties=properties
        )
        parameters.properties.enabled_for_deployment = True
        parameters.properties.enabled_for_disk_encryption = True
        parameters.properties.enabled_for_template_deployment = True
        return parameters


# pylint: enable=too-many-instance-attributes


def _user_oid(token) -> str:
    """
    Return the user Object ID.

    Returns
    -------
    str
        User OID.

    """
    data = _get_parsed_token_data(token)
    return data.get("oid")


def _get_parsed_token_data(token) -> Any:
    tok_data = token
    tok_data = tok_data.split(".")[1]
    tok_data += "=" * ((4 - len(tok_data) % 4) % 4)
    return json.loads(base64.b64decode(tok_data))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
msticpy - Jupyter and Python Tools for InfoSec.

-----------------------------------------------

Requires Python 3.6 or later.

To quickly import common modules into a notebook run:

>>> from msticpy import init_notebook
>>> init_notebook(globals())

For more options:
>>> help(init_notebook)

-----------------------------------------------

Full documentation is available at:
    https://msticpy.readthedocs.io

GitHub repo:
    https://github.com/microsoft/msticpy

"""

import os

# flake8: noqa: F403
from .nbtools.nbinit import init_notebook, current_providers
from .common import pkg_config as settings
from .common.check_version import check_version
from . import sectools
from . import nbtools
from . import data
from .config.mp_config_edit import MpConfigEdit, MpConfigFile
from ._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen, Pete Bryan, Ashwin Patil"

if not os.environ.get("KQLMAGIC_EXTRAS_REQUIRES"):
    os.environ["KQLMAGIC_EXTRAS_REQUIRES"] = "jupyter-basic"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Morph Charts  class."""
import json
from pathlib import Path
import yaml
from IPython.display import IFrame
import pandas as pd
from ..common.exceptions import MsticpyException

from .._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"

_CHART_FOLDER = "morph_charts"


class MorphCharts:
    """Create Morph Charts package data and render Morph Charts site."""

    def __init__(self):
        """Create object and populate charts container."""
        self.charts = _get_charts(_CHART_FOLDER)

    def display(self, data: pd.DataFrame, chart_name: str) -> IFrame:
        """
        Prepare package data and display MorphChart in an IFrame.

        Parameters
        ----------
        data: pd.DataFrame:
            A DataFrame of data for the morphchart to plot.

        chart_name: str:
            The name of the Morph Chart to plot.

        """
        # Check input data is correct format and that the chart being requested exists
        if not isinstance(data, pd.DataFrame):
            raise MsticpyException("Data provided must be in pandas.DataFrame format")

        if chart_name not in self.charts:
            raise MsticpyException(
                f"{chart_name} is not a vaid chart. Run list_charts() to see avaliable charts"  # pylint: disable=line-too-long
            )

        # Create description file with length of our data set
        description_dict = self.charts[chart_name]["DescriptionFile"]
        description_dict["tables"][0]["rows"] = len(data)
        # Create output folder for package files
        out_path = Path.cwd().joinpath(*["morphchart_package", "description.json"])
        Path.mkdir(Path.cwd().joinpath("morphchart_package"), exist_ok=True)
        # Write description file
        with open(out_path, "w", encoding="utf-8") as morph_file:
            json.dump(description_dict, morph_file)
        # Write dataset to query_data csv
        data_out_path = out_path = Path.cwd().joinpath(
            *["morphchart_package", "query_data.csv"]
        )
        data.to_csv(data_out_path, index=False)
        # Display Morph Charts in IFrame with instructions
        print(
            f"Navigate to {Path.cwd().joinpath('morphchart_package')} and upload the files below"
        )
        print("Charts provided by http://morphcharts.com/")
        return IFrame("http://morphcharts.com/designer.html", "100%", "600px")

    def list_charts(self):
        """Get a list of avaliable charts."""
        for key, _ in self.charts.items():
            print(key)

    def get_chart_details(self, chart_name):
        """
        Get description for a chart.

        Parameters
        ----------
        chart_name: str:
            The name of the chart you get description for.

        """
        try:
            print(
                chart_name,
                ":",
                "\n",
                self.charts[chart_name]["Description"],
                "\n",
                "Query: ",
                self.charts[chart_name]["Query"],
            )
        except KeyError as key_err:
            raise KeyError(f"Unknown chart {chart_name}") from key_err

    def search_charts(self, keyword):
        """
        Search for charts that match a keyword.

        Parameters
        ----------
        keyword: str:
            The keyword to search charts for.

        """
        for key, value in self.charts.items():
            if keyword.casefold() in [tag.casefold() for tag in value["Tags"]]:
                print(key, ":", "\n", value["Description"])
            elif keyword.casefold() in [
                word.casefold() for word in value["Description"].split()
            ]:
                print(key, ":", "\n", value["Description"])
            else:
                print("No matching charts found")


def _get_charts(path: str = "morph_charts") -> dict:
    """
    Return dictionary of yaml files found in the Morph Charts folder.

    Parameters
    ----------
    path : str
        The source path to search in.

    Returns
    -------
    Dict
        Details of the chart files

    """
    full_path = Path(__file__).parent.parent.joinpath("data").joinpath(path)
    file_glob = Path(full_path).glob("*.yaml")
    chart_files = [file_path for file_path in file_glob if file_path.is_file()]
    chart_details = {}
    for chart in chart_files:
        with open(chart, "r", encoding="utf-8") as chart_data:
            details = yaml.safe_load(chart_data)
        try:
            chart_details.update(
                {
                    details["Name"]: {
                        "Description": details["Description"],
                        "Query": details["Query"],
                        "Tags": details["Tags"],
                        "DescriptionFile": details["DescriptionFile"],
                    }
                }
            )
        except KeyError as key_err:
            raise Exception(
                f"{chart} description does not appear to be in the correct format."
            ) from key_err

    return chart_details



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
ip_utils - IP Address functions.

Contains a series of functions required to manipulate and enrich IP Address data
to assist investigations.

Designed to support any data source containing IP address entity.

"""
import ipaddress
from functools import lru_cache
from typing import Callable, List, Optional, Set, Tuple

import pandas as pd
from deprecated.sphinx import deprecated
from ipwhois import (
    ASNRegistryError,
    HostLookupError,
    HTTPLookupError,
    HTTPRateLimitError,
    IPWhois,
    WhoisLookupError,
    WhoisRateLimitError,
)

from .._version import VERSION
from ..common.utility import arg_to_list, export
from ..datamodel.entities import GeoLocation, IpAddress
from .geoip import GeoLiteLookup

__version__ = VERSION
__author__ = "Ashwin Patil"


class Error(Exception):
    """Base class for other exceptions."""


class DataError(Error):
    """Raised when thereis a data input error."""


def _get_geolite_lookup() -> Callable:
    """Closure for instantiating GeoLiteLookup."""
    geo_ip = None

    def _get_geo_ip(**kwargs) -> GeoLiteLookup:
        nonlocal geo_ip
        if geo_ip is None:
            geo_ip = GeoLiteLookup(**kwargs)
        return geo_ip

    return _get_geo_ip


_GET_IP_LOOKUP = _get_geolite_lookup()


def convert_to_ip_entities(  # noqa: MC0001
    ip_str: Optional[str] = None,
    data: Optional[pd.DataFrame] = None,
    ip_col: Optional[str] = None,
    geo_lookup: bool = True,
) -> List[IpAddress]:
    """
    Take in an IP Address string and converts it to an IP Entity.

    Parameters
    ----------
    ip_str : str
        A string with a single IP Address or multiple addresses
        delimited by comma or space
    data : pd.DataFrame
        Use DataFrame as input
    ip_col : str
        Column containing IP addresses
    geo_lookup : bool
        If true, do geolocation lookup on IPs,
        by default, True

    Returns
    -------
    List
        The populated IP entities including address and geo-location

    Raises
    ------
    ValueError
        If neither ip_string or data/column provided as input

    """
    ip_entities: List[IpAddress] = []
    all_ips: Set[str] = set()

    if geo_lookup:
        ip_lookup = _GET_IP_LOOKUP()
    if ip_str:
        addrs = arg_to_list(ip_str)
    elif data is not None and ip_col:
        addrs = data[ip_col].values
    else:
        raise ValueError("Must specify either ip_str or data + ip_col parameters.")

    for addr in addrs:
        if isinstance(addr, list):
            ip_list = set(addr)
        elif isinstance(addr, str) and "," in addr:
            ip_list = {ip.strip() for ip in addr.split(",")}
        else:
            ip_list = {addr}
        ip_list = ip_list - all_ips  # remove IP addresses we've seen
        ip_entities.extend(IpAddress(Address=ip) for ip in ip_list)
        all_ips |= ip_list
        if geo_lookup:
            for ip_ent in ip_entities:
                try:
                    ip_lookup.lookup_ip(ip_entity=ip_ent)
                except DataError:
                    pass
    return ip_entities


@export  # noqa: MC0001
# pylint: disable=too-many-return-statements, invalid-name
def get_ip_type(ip: str = None, ip_str: str = None) -> str:  # noqa: MC0001
    """
    Validate value is an IP address and determine IPType category.

    (IPAddress category is e.g. Private/Public/Multicast).

    Parameters
    ----------
    ip : str
        The string of the IP Address
    ip_str : str
        The string of the IP Address - alias for `ip`

    Returns
    -------
    str
        Returns ip type string using ip address module

    """
    ip_str = ip or ip_str
    if not ip_str:
        raise ValueError("'ip' or 'ip_str' value must be specified")
    try:
        ipaddress.ip_address(ip_str)
    except ValueError:
        print(f"{ip_str} does not appear to be an IPv4 or IPv6 address")
    else:
        if ipaddress.ip_address(ip_str).is_multicast:
            return "Multicast"
        if ipaddress.ip_address(ip_str).is_global:
            return "Public"
        if ipaddress.ip_address(ip_str).is_loopback:
            return "Loopback"
        if ipaddress.ip_address(ip_str).is_link_local:
            return "Link Local"
        if ipaddress.ip_address(ip_str).is_unspecified:
            return "Unspecified"
        if ipaddress.ip_address(ip_str).is_private:
            return "Private"
        if ipaddress.ip_address(ip_str).is_reserved:
            return "Reserved"

    return "Unspecified"


# pylint: enable=too-many-return-statements


# pylint: disable=invalid-name
@lru_cache(maxsize=1024)
def get_whois_info(
    ip: str = None, show_progress: bool = False, **kwargs
) -> Tuple[str, dict]:
    """
    Retrieve whois ASN information for given IP address using IPWhois python package.

    Parameters
    ----------
    ip : str
        IP Address to look up.
    ip_str : str
        alias for `ip`.
    show_progress : bool, optional
        Show progress for each query, by default False

    Returns
    -------
    IP
        Details of the IP data collected

    Notes
    -----
    This function uses the Python functools lru_cache and
    will return answers from the cache for previously queried
    IP addresses.

    """
    ip_str = ip or kwargs.get("ip_str")
    if not ip_str:
        raise ValueError("'ip' or 'ip_str' value must be specified")
    ip_type = get_ip_type(ip_str)
    if ip_type == "Public":
        try:
            whois = IPWhois(ip_str)
            whois_result = whois.lookup_whois()
            if show_progress:
                print(".", end="")
            return whois_result["asn_description"], whois_result
        except (
            HTTPLookupError,
            HTTPRateLimitError,
            HostLookupError,
            WhoisLookupError,
            WhoisRateLimitError,
            ASNRegistryError,
        ) as err:
            return f"Error during lookup of {ip_str} {type(err)}", {}
    return f"No ASN Information for IP type: {ip_type}", {}


# pylint: enable=invalid-name


def get_whois_df(
    data: pd.DataFrame,
    ip_column: str,
    all_columns: bool = False,
    asn_col: str = "AsnDescription",
    whois_col: Optional[str] = None,
    show_progress: bool = False,
) -> pd.DataFrame:
    """
    Retrieve Whois ASN information for DataFrame of IP Addresses.

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame
    ip_column : str
        Column name of IP Address to look up.
    all_columns:
        Expand all whois data to columns.
    asn_col : str, optional
        Name of the output column for ASN description,
        by default "ASNDescription".
        Ignored if `all_columns` is True.
    whois_col : str, optional
        Name of the output column for full whois data,
        by default "WhoIsData"
        Ignored if `all_columns` is True.
    show_progress : bool, optional
        Show progress for each query, by default False

    Returns
    -------
    pd.DataFrame
        Output DataFrame with results in added columns.

    """
    if all_columns:
        return data.apply(
            lambda x: get_whois_info(x[ip_column], show_progress=show_progress)[1],
            axis=1,
            result_type="expand",
        )
    data = data.copy()
    if whois_col is not None:
        data[[asn_col, whois_col]] = data.apply(
            lambda x: get_whois_info(x[ip_column], show_progress=show_progress),
            axis=1,
            result_type="expand",
        )
    else:
        data[asn_col] = data.apply(
            lambda x: get_whois_info(x[ip_column], show_progress=show_progress)[0],
            axis=1,
        )
    return data


# pylint: disable=too-few-public-methods
@pd.api.extensions.register_dataframe_accessor("mp_whois")
class IpWhoisAccessor:
    """Pandas api extension for IP Whois lookup."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj

    def lookup(self, ip_column, **kwargs):
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        ip_column : str
            Column name of IP Address to look up.

        Other Parameters
        ----------------
        asn_col : str, optional
            Name of the output column for ASN description,
            by default "ASNDescription"
        whois_col : str, optional
            Name of the output column for full whois data,
            by default "WhoIsData"
        show_progress : bool, optional
            Show progress for each query, by default False

        Returns
        -------
        pd.DataFrame
            Output DataFrame with results in added columns.

        """
        return get_whois_df(data=self._df, ip_column=ip_column, **kwargs)


@deprecated("Will be removed in a future version", version="1.4.0")
def create_ip_record(
    heartbeat_df: pd.DataFrame, az_net_df: pd.DataFrame = None
) -> IpAddress:
    """
    Generate ip_entity record for provided IP value.

    Parameters
    ----------
    heartbeat_df : pd.DataFrame
        A dataframe of heartbeat data for the host
    az_net_df : pd.DataFrame
        Option dataframe of Azure network data for the host

    Returns
    -------
    IP
        Details of the IP data collected

    """
    ip_entity = IpAddress()

    # Produce ip_entity record using available dataframes
    ip_hb = heartbeat_df.iloc[0]
    ip_entity.Address = ip_hb["ComputerIP"]
    ip_entity.hostname = ip_hb["Computer"]  # type: ignore
    ip_entity.SourceComputerId = ip_hb["SourceComputerId"]  # type: ignore
    ip_entity.OSType = ip_hb["OSType"]  # type: ignore
    ip_entity.OSName = ip_hb["OSName"]  # type: ignore
    ip_entity.OSVMajorersion = ip_hb["OSMajorVersion"]  # type: ignore
    ip_entity.OSVMinorVersion = ip_hb["OSMinorVersion"]  # type: ignore
    ip_entity.ComputerEnvironment = ip_hb["ComputerEnvironment"]  # type: ignore
    ip_entity.OmsSolutions = [  # type: ignore
        sol.strip() for sol in ip_hb["Solutions"].split(",")
    ]
    ip_entity.VMUUID = ip_hb["VMUUID"]  # type: ignore
    ip_entity.SubscriptionId = ip_hb["SubscriptionId"]  # type: ignore
    geoloc_entity = GeoLocation()  # type: ignore
    geoloc_entity.CountryName = ip_hb["RemoteIPCountry"]  # type: ignore
    geoloc_entity.Longitude = ip_hb["RemoteIPLongitude"]  # type: ignore
    geoloc_entity.Latitude = ip_hb["RemoteIPLatitude"]  # type: ignore
    ip_entity.Location = geoloc_entity  # type: ignore

    # If Azure network data present add this to host record
    if az_net_df is not None and not az_net_df.empty:
        if len(az_net_df) == 1:
            priv_addr_str = az_net_df["PrivateIPAddresses"].loc[0]
            ip_entity["private_ips"] = convert_to_ip_entities(priv_addr_str)
            pub_addr_str = az_net_df["PublicIPAddresses"].loc[0]
            ip_entity["public_ips"] = convert_to_ip_entities(pub_addr_str)
        else:
            if "private_ips" not in ip_entity:
                ip_entity["private_ips"] = []
            if "public_ips" not in ip_entity:
                ip_entity["public_ips"] = []

    return ip_entity



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""FileHash Entity class."""
from typing import Any, Mapping

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .entity_enums import Algorithm

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class FileHash(Entity):
    """
    File Hash class.

    Attributes
    ----------
    Algorithm : Algorithm
        FileHash Algorithm
    Value : str
        FileHash Value

    """

    ID_PROPERTIES = ["Value"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Algorithm: Algorithm = Algorithm.Unknown
        self.Value: str = ""
        super().__init__(src_entity=src_entity, **kwargs)
        if src_event is not None:
            self._create_from_event(src_event)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Algorithm}: {self.Value}"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Value

    def _create_from_event(self, src_event):
        self.Algorithm = src_event["Algorithm"]
        self.Value = src_event["HashValue"]

    _entity_schema = {
        # The hash algorithm (type System.String)
        "Algorithm": "Algorithm",
        # Value (type System.String)
        "Value": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Entity graph property."""
from typing import Union

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

# flake8: noqa: F821
# "Entity" cannot be imported due to circular reference


# Future - will replace entity graph creation with property factory
def graph_property(
    name: str, prop_type: Union[type, str], edge_name: str = None
) -> property:
    """Property factory for graph_property."""
    storage_name = f"_{name}"
    edge_attrs = {"name": edge_name or name}
    prop_doc = f"Get, set or delete {name} property."
    if not isinstance(prop_type, type) or not prop_type == "self":
        raise TypeError(
            f"{prop_type} is not resolvable.",
            "'prop_type' must be a type of Entity or the string 'self'",
        )

    def prop_getter(self: "Entity") -> "Entity":  # type: ignore
        """Return property value."""
        return getattr(self, storage_name, None)

    def prop_setter(self: "Entity", value: "Entity"):  # type: ignore
        """Set property value and add graph edge."""
        nonlocal prop_type
        if prop_type == "self":
            prop_type = self.__class__
        if not isinstance(prop_type, type) or not isinstance(value, prop_type):
            raise TypeError(
                f"Cannot assign {type(value)} to property {name}.",
                f"Property must be of type {prop_type}",
            )
        if hasattr(self, storage_name):
            delattr(self, storage_name)
            for edge in self.edges:
                if edge.attrs.get("name") == name:
                    self.edges.remove(edge)
                    break

        setattr(self, storage_name, value)
        self.add_edge(target=value, edge_attrs=edge_attrs)

    def prop_del(self):
        """Property deleter."""
        if not hasattr(self, storage_name):
            return
        for edge in self.edges:
            if edge.attrs.get("name") == name:
                self.edges.remove(edge)
                break
        delattr(self, storage_name)

    return property(prop_getter, prop_setter, prop_del, prop_doc)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Deprecated placeholder for eventcluster.py."""
import warnings
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


# flake8: noqa: F403, F401
# pylint: disable=wildcard-import, unused-wildcard-import, unused-import
from ..analysis.eventcluster import *

WARN_MSSG = (
    "This module has moved to msticpy.analysis.eventcluster\n"
    + "Please change your import to reflect this new location."
)
warnings.warn(WARN_MSSG, category=DeprecationWarning)



# -------------------------------------------------------------------------
# Copyright (c) SecureWorks. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure Resource Graph Driver class."""
from typing import Any, Tuple, Union
import warnings

import pandas as pd
from pandas.core.frame import DataFrame

from .driver_base import DriverBase, QuerySource
from ..._version import VERSION
from ...common.utility import export
from ...common.exceptions import (
    MsticpyNotConnectedError,
    MsticpyImportExtraError,
)
from ...common.azure_auth import az_connect, AzureCloudConfig, only_interactive_cred

# pylint: disable=wrong-import-order
from azure.mgmt.subscription import SubscriptionClient

try:
    from azure.mgmt.resourcegraph import ResourceGraphClient
    from azure.mgmt.resourcegraph.models import (
        ResultTruncated,
        QueryRequest,
        QueryRequestOptions,
        QueryResponse,
        ResultFormat,
    )
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without azure-mgmt-resourcegraph installed",
        title="Error importing azure-mgmt-resourcegraph",
        extra="azure",
    ) from imp_err
# pylint: enable=wrong-import-order

__version__ = VERSION
__author__ = "Ryan Cobb"


@export
class ResourceGraphDriver(DriverBase):
    """Driver to connect and query from Azure Resource Graph."""

    def __init__(self, **kwargs):
        """Instantiate Azure Resource Graph Driver."""
        super().__init__()
        self.client = None
        self.sub_client = None
        self.subscription_ids = None
        self._connected = False
        self._debug = kwargs.get("debug", False)
        self.az_cloud_config = AzureCloudConfig(cloud=kwargs.get("cloud"))

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to Azure Resource Graph via Azure SDK.

        Parameters
        ----------
        connection_str : Optional[str], optional
            Not used.

        Other Parameters
        ----------------
        kwargs :
            Connection parameters can be supplied as keyword parameters.

        Notes
        -----
        Default configuration is read from the DataProviders/AzureCLI
        section of msticpyconfig.yaml, if available.

        """
        auth_methods = kwargs.get("auth_methods")
        auth_methods = auth_methods or self.az_cloud_config.auth_methods
        silent = kwargs.get("silent", True)

        credentials = az_connect(auth_methods=auth_methods, silent=silent)
        if only_interactive_cred(credentials.modern):
            print("Check your default browser for interactive sign-in prompt.")
        self.client = ResourceGraphClient(
            credential=credentials.modern,
            base_url=self.az_cloud_config.endpoints.resource_manager,
            credential_scopes=[self.az_cloud_config.token_uri],
        )
        self.sub_client = SubscriptionClient(
            credential=credentials.modern,
            base_url=self.az_cloud_config.endpoints.resource_manager,
            credential_scopes=[self.az_cloud_config.token_uri],
        )
        self.subscription_ids = [
            sub.subscription_id
            for sub in self.sub_client.subscriptions.list()  # type: ignore
        ]

        self._connected = True
        self._loaded = True

        print("Connected")

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute Resource Graph query and retrieve results.

        Parameters
        ----------
        query : str
            KQL query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            count

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        del query_source
        result_df, result = self.query_with_results(query, **kwargs)
        if isinstance(result_df, DataFrame) and not result_df.empty:
            return result_df

        return result

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            Query to execute against Resource Graph

        Returns
        -------
        Union[pd.DataFrame,Any]
            A DataFrame (if successful) or
            the underlying provider result if an error occurs.

        """
        if not self.connected:
            self.connect()
        if not self.connected:
            raise MsticpyNotConnectedError(
                "Source is not connected. ", "Please call connect() and retry."
            )

        result_truncated = False

        top = kwargs.get("top", 1000)

        request_options = QueryRequestOptions(
            top=top,
            result_format=ResultFormat.object_array,
        )

        request = QueryRequest(
            query=query,
            subscriptions=self.subscription_ids,
            options=request_options,
        )

        response = self.client.resources(request)  # type: QueryResponse

        # Pagination logic adapted from azure-cli-extensions
        # https://github.com/Azure/azure-cli-extensions/blob/8dade2f6fe28803d0fbdb1700c3ab4e4d71e5318/src/resource-graph/azext_resourcegraph/custom.py#L75

        if response.result_truncated == ResultTruncated.true:
            result_truncated = True

        if result_truncated and top is not None and len(response.data) < top:
            warnings.warn(
                "Unable to paginate the results of the query. "
                "Some resources may be missing from the results. "
                "To rewrite the query and enable paging, "
                "see the docs for an example: https://aka.ms/arg-results-truncated",
            )

        return pd.json_normalize(response.data), response



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Entity Graph classes."""
from typing import Any, Dict, Set, Optional

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


class Node:
    """Entity node."""

    def __init__(self):
        """Initialize the node."""
        self.edges: Set["Edge"] = set()

    def add_edge(self, target: "Node", edge_attrs: Optional[Dict[str, Any]] = None):
        """
        Add an edge between self and target.

        Parameters
        ----------
        target : Node
            Target node.
        edge_attrs : Optional[Dict[str, Any]], optional
            Attributes to assign to new edge, by default None

        """
        edge = Edge(self, target, edge_attrs) if edge_attrs else Edge(self, target)
        if not self.has_edge(target):
            self.edges.add(edge)
        if not target.has_edge(self):
            target.edges.add(edge)

    def has_edge(self, other):
        """Return True if node has an edge with `other`."""
        return any(edge for edge in self.edges if other in (edge.target, edge.source))


class Edge:
    """Entity edge class."""

    def __init__(self, source: Node, target: Node, attrs: Dict[str, Any] = None):
        """
        Create a new edge between `source` and `target`.

        Parameters
        ----------
        source : Node
            Source node.
        target : Node
            Target node.
        attrs : Dict[str, Any], optional
            Dictionary of name/value edge attributes, by default None

        """
        self.source: Node = source
        self.target: Node = target

        self.attrs: Dict[str, Any] = attrs or {}

    def add_attr(self, name: str, value: Any):
        """Add an edge attribute."""
        self.attrs[name] = value

    def __str__(self):
        """Return string representation of edge."""
        return self.attrs.get("name", "unnamed-edge")

    def __repr__(self):
        """Return full repr of edge."""
        other_attrs = [
            f"{name}='{val}'" for name, val in self.attrs.items() if name != "name"
        ]
        if not other_attrs:
            return f"Edge(name={str(self)})"
        return f"Edge(name={str(self)}, {', '.join(other_attrs)})"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Threatintelligence Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class Threatintelligence(Entity):
    """
    Threatintelligence Entity class.

    Attributes
    ----------
    ProviderName : str
        Threatintelligence ProviderName
    ThreatType : str
        Threatintelligence ThreatType
    ThreatName : str
        Threatintelligence ThreatName
    Confidence : str
        Threatintelligence Confidence
    ReportLink : str
        Threatintelligence ReportLink
    ThreatDescription : str
        Threatintelligence ThreatDescription

    """

    ID_PROPERTIES = ["ProviderName", "ThreatName", "ReportLink"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

            :param src_entity: instantiate entity using properties of src entity
            :param kwargs: key-value pair representation of entity
        """
        self.ProviderName: Optional[str] = None
        self.ThreatType: Optional[str] = None
        self.ThreatName: Optional[str] = None
        self.Confidence: Optional[str] = None
        self.ReportLink: Optional[str] = None
        self.ThreatDescription: Optional[str] = None
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.ThreatName} ({self.ThreatType})"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.ThreatName or self.__class__.__name__

    _entity_schema = {
        # String Name of the provider from whom this
        # Threat Intelligence information was received
        "ProviderName": None,
        "ThreatType": None,
        "ThreatName": None,
        "Confidence": None,
        "ReportLink": None,
        "ThreatDescription": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }
